"""
Unit tests for AWS KMS Decryption Renderer.
"""
import pytest

import salt.exceptions
import salt.renderers.aws_kms as aws_kms
from tests.support.mock import MagicMock, patch

try:
    import botocore.exceptions
    import botocore.session
    import botocore.stub

    HAS_BOTOCORE = True
except ImportError:
    HAS_BOTOCORE = False

try:
    import cryptography.fernet as fernet

    HAS_FERNET = True
except ImportError:
    HAS_FERNET = False

pytestmark = [
    pytest.mark.skipif(
        HAS_BOTOCORE is False, reason="Unable to import botocore libraries"
    )
]


@pytest.fixture
def plaintext_secret():
    return "Us): more salt."


@pytest.fixture
def encrypted_data_key():
    return "encrypted-data-key"


@pytest.fixture
def plaintext_data_key():
    return b"plaintext-data-key"


@pytest.fixture
def base64_data_key():
    return b"cGxhaW50ZXh0LWRhdGEta2V5"


@pytest.fixture
def aws_profile():
    return "test-profile"


@pytest.fixture
def region_name():
    return "us-test-1"


@pytest.fixture
def configure_loader_modules():
    return {aws_kms: {}}


def test__cfg_data_key(encrypted_data_key):
    """
    _cfg_data_key returns the aws_kms:data_key from configuration.
    """
    config = {"aws_kms": {"data_key": encrypted_data_key}}
    with patch.dict(aws_kms.__salt__, {"config.get": config.get}):
        assert (
            aws_kms._cfg_data_key() == encrypted_data_key
        ), "_cfg_data_key did not return the data key configured in __salt__."
    with patch.dict(aws_kms.__opts__, config):
        assert (
            aws_kms._cfg_data_key() == encrypted_data_key
        ), "_cfg_data_key did not return the data key configured in __opts__."


def test__cfg_data_key_no_key():
    """
    When no aws_kms:data_key is configured,
    calling _cfg_data_key should raise a SaltConfigurationError
    """
    pytest.raises(salt.exceptions.SaltConfigurationError, aws_kms._cfg_data_key)


def test__session_profile(aws_profile):
    """
    _session instantiates boto3.Session with the configured profile_name
    """
    with patch.object(aws_kms, "_cfg", lambda k: aws_profile):
        with patch("boto3.Session") as session:
            aws_kms._session()
            session.assert_called_with(profile_name=aws_profile)


def test__session_noprofile(aws_profile):
    """
    _session raises a SaltConfigurationError
    when boto3 raises botocore.exceptions.ProfileNotFound.
    """
    with patch("boto3.Session") as session:
        session.side_effect = botocore.exceptions.ProfileNotFound(profile=aws_profile)
        pytest.raises(salt.exceptions.SaltConfigurationError, aws_kms._session)


def test__session_noregion():
    """
    _session raises a SaltConfigurationError
    when boto3 raises botocore.exceptions.NoRegionError
    """
    with patch("boto3.Session") as session:
        session.side_effect = botocore.exceptions.NoRegionError
        pytest.raises(salt.exceptions.SaltConfigurationError, aws_kms._session)


def test__kms():
    """
    _kms calls boto3.Session.client with 'kms' as its only argument.
    """
    with patch("boto3.Session.client") as client:
        aws_kms._kms()
        client.assert_called_with("kms")


def test__kms_noregion():
    """
    _kms raises a SaltConfigurationError
    when boto3 raises a NoRegionError.
    """
    with patch("boto3.Session") as session:
        session.side_effect = botocore.exceptions.NoRegionError
        pytest.raises(salt.exceptions.SaltConfigurationError, aws_kms._kms)


def test__api_decrypt(encrypted_data_key):
    """
    _api_decrypt_response calls kms.decrypt with the
    configured data key as the CiphertextBlob kwarg.
    """
    kms_client = MagicMock()
    with patch.object(aws_kms, "_kms") as kms_getter:
        kms_getter.return_value = kms_client
        with patch.object(aws_kms, "_cfg_data_key", lambda: encrypted_data_key):
            aws_kms._api_decrypt()
            kms_client.decrypt.assert_called_with(CiphertextBlob=encrypted_data_key)


def test__api_decrypt_badkey(encrypted_data_key):
    """
    _api_decrypt_response raises SaltConfigurationError
    when kms.decrypt raises a botocore.exceptions.ClientError
    with an error_code of 'InvalidCiphertextException'.
    """
    kms_client = MagicMock()
    kms_client.decrypt.side_effect = botocore.exceptions.ClientError(
        error_response={"Error": {"Code": "InvalidCiphertextException"}},
        operation_name="Decrypt",
    )
    with patch.object(aws_kms, "_kms") as kms_getter:
        kms_getter.return_value = kms_client
        with patch.object(aws_kms, "_cfg_data_key", lambda: encrypted_data_key):
            pytest.raises(salt.exceptions.SaltConfigurationError, aws_kms._api_decrypt)


def test__plaintext_data_key(plaintext_data_key):
    """
    _plaintext_data_key returns the 'Plaintext' value from the response.
    It caches the response and only calls _api_decrypt exactly once.
    """
    with patch.object(
        aws_kms,
        "_api_decrypt",
        return_value={"KeyId": "key-id", "Plaintext": plaintext_data_key},
    ) as api_decrypt:
        assert aws_kms._plaintext_data_key() == plaintext_data_key
        aws_kms._plaintext_data_key()
        api_decrypt.assert_called_once()


def test__base64_plaintext_data_key(plaintext_data_key, base64_data_key):
    """
    _base64_plaintext_data_key returns the urlsafe base64 encoded plain text data key.
    """
    with patch.object(aws_kms, "_plaintext_data_key", return_value=plaintext_data_key):
        assert aws_kms._base64_plaintext_data_key() == base64_data_key


@pytest.mark.skipif(HAS_FERNET is False, reason="Failed to import cryptography.fernet")
def test__decrypt_ciphertext(plaintext_secret):
    """
    test _decrypt_ciphertext
    """
    test_key = fernet.Fernet.generate_key()
    crypted = fernet.Fernet(test_key).encrypt(plaintext_secret.encode())
    with patch.object(aws_kms, "_base64_plaintext_data_key", return_value=test_key):
        assert aws_kms._decrypt_ciphertext(crypted) == plaintext_secret


@pytest.mark.skipif(HAS_FERNET is False, reason="Failed to import cryptography.fernet")
def test__decrypt_object(plaintext_secret):
    """
    Test _decrypt_object
    """
    test_key = fernet.Fernet.generate_key()
    crypted = fernet.Fernet(test_key).encrypt(plaintext_secret.encode())
    secret_map = {"secret": plaintext_secret}
    crypted_map = {"secret": crypted}

    secret_list = [plaintext_secret]
    crypted_list = [crypted]

    with patch.object(aws_kms, "_base64_plaintext_data_key", return_value=test_key):
        assert aws_kms._decrypt_object(plaintext_secret) == plaintext_secret
        assert aws_kms._decrypt_object(crypted) == plaintext_secret
        assert aws_kms._decrypt_object(crypted_map) == secret_map
        assert aws_kms._decrypt_object(crypted_list) == secret_list
        assert aws_kms._decrypt_object(None) is None


@pytest.mark.skipif(HAS_FERNET is False, reason="Failed to import cryptography.fernet")
def test_render(plaintext_secret):
    """
    Test that we can decrypt some data.
    """
    test_key = fernet.Fernet.generate_key()
    crypted = fernet.Fernet(test_key).encrypt(plaintext_secret.encode())
    with patch.object(aws_kms, "_base64_plaintext_data_key", return_value=test_key):
        assert aws_kms.render(crypted) == plaintext_secret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import logging

import pytest

import salt.modules.rabbitmq as rabbitmq
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {rabbitmq: {"__context__": {"rabbitmqctl": None, "rabbitmq-plugins": None}}}


# 'list_users_rabbitmq2' function tests: 1
def test_list_users_rabbitmq2():
    """
    Test if it return a list of users based off of rabbitmqctl user_list.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                "Listing users ...\nguest\t[administrator,"
                " user]\njustAnAdmin\t[administrator]\n"
            ),
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_users() == {
            "guest": ["administrator", "user"],
            "justAnAdmin": ["administrator"],
        }


# 'list_users_rabbitmq3' function tests: 1
def test_list_users_rabbitmq3():
    """
    Test if it return a list of users based off of rabbitmqctl user_list.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "guest\t[administrator user]\r\nother\t[a b]\r\n",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_users() == {
            "guest": ["administrator", "user"],
            "other": ["a", "b"],
        }


# 'list_users_with_warning_rabbitmq2' function tests: 1
def test_list_users_with_warning_rabbitmq2():
    """
    Test if having a leading WARNING returns the user_list anyway.
    """
    rtn_stdout = "\n".join(
        [
            "WARNING: ignoring /etc/rabbitmq/rabbitmq.conf -- location has moved to"
            " /etc/rabbitmq/rabbitmq-env.conf",
            "Listing users ...",
            "guest\t[administrator, user]\n",
        ]
    )
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": rtn_stdout, "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_users() == {"guest": ["administrator", "user"]}


# 'list_users_with_warning_rabbitmq3' function tests: 1
def test_list_users_with_warning_rabbitmq3():
    """
    Test if having a leading WARNING returns the user_list anyway.
    """
    rtn_stdout = "\n".join(
        [
            "WARNING: ignoring /etc/rabbitmq/rabbitmq.conf -- location has moved to"
            " /etc/rabbitmq/rabbitmq-env.conf",
            "Listing users ...",
            "guest\t[administrator user]\n",
        ]
    )
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": rtn_stdout, "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_users() == {"guest": ["administrator", "user"]}


# 'list_vhosts' function tests: 2
def test_list_vhosts():
    """
    Test if it return a list of vhost based on rabbitmqctl list_vhosts.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "/\nsaltstack\n...", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_vhosts() == ["/", "saltstack", "..."]


def test_list_vhosts_with_warning():
    """
    Test if it return a list of vhost based on rabbitmqctl list_vhosts even with a leading WARNING.
    """
    rtn_stdout = "\n".join(
        [
            "WARNING: ignoring /etc/rabbitmq/rabbitmq.conf -- location has moved to"
            " /etc/rabbitmq/rabbitmq-env.conf",
            "Listing users ...",
            "/",
            "saltstack",
            "...\n",
        ]
    )
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": rtn_stdout, "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_vhosts() == ["/", "saltstack", "..."]


# 'user_exists' function tests: 2
def test_user_exists():
    """
    Test whether a given rabbitmq-internal user exists based
    on rabbitmqctl list_users.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "Listing users ...\nsaltstack\t[administrator]\n...done",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.user_exists("saltstack")


def test_user_exists_negative():
    """
    Negative test of whether rabbitmq-internal user exists based
    on rabbitmqctl list_users.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "Listing users ...\nsaltstack\t[administrator]\n...done",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert not rabbitmq.user_exists("salt")


# 'vhost_exists' function tests: 2
def test_vhost_exists():
    """
    Test if it return whether the vhost exists based
    on rabbitmqctl list_vhosts.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "Listing vhosts ...\nsaltstack",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.vhost_exists("saltstack")


def test_vhost_exists_negative():
    """
    Test if it return whether the vhost exists based
    on rabbitmqctl list_vhosts.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "Listing vhosts ...\nsaltstack",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert not rabbitmq.vhost_exists("salt")


# 'add_user' function tests: 1
def test_add_user():
    """
    Test if it add a rabbitMQ user via rabbitmqctl
    user_add <user> <password>
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.add_user("saltstack") == {"Added": "saltstack"}

    mock_run = MagicMock(return_value="Error")
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        with patch.object(
            rabbitmq,
            "clear_password",
            return_value={"Error": "Error", "retcode": 1},
        ):
            pytest.raises(CommandExecutionError, rabbitmq.add_user, "saltstack")


# 'delete_user' function tests: 1
def test_delete_user():
    """
    Test if it deletes a user via rabbitmqctl delete_user.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.delete_user("saltstack") == {"Deleted": "saltstack"}


# 'check_password' function tests: 2
def test_check_password_lt_38():
    """
    Test if it checks a user's password for RabbitMQ less than v3.8.
    """
    mock_run = MagicMock(return_value='{rabbit,"RabbitMQ","3.5.7"}')
    mock_run2 = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": 'Authenticating user "saltstack" ...\nSuccess',
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run2}):
        assert rabbitmq.check_password("saltstack", "salt@123")


def test_check_password_gt_38():
    """
    Test if it checks a user's password for RabbitMQ greater than v3.8.
    """
    mock_run = MagicMock(return_value="RabbitMQ version: 3.8.3")
    mock_run2 = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": 'Authenticating user "saltstack" ...\nSuccess',
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run2}):
        assert rabbitmq.check_password("saltstack", "salt@123")


# 'change_password' function tests: 1
def test_change_password():
    """
    Test if it changes a user's password.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.change_password("saltstack", "salt@123") == {
            "Password Changed": "saltstack"
        }


# 'clear_password' function tests: 1
def test_clear_password():
    """
    Test if it removes a user's password.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.clear_password("saltstack") == {"Password Cleared": "saltstack"}


# 'add_vhost' function tests: 1
def test_add_vhost():
    """
    Test if it adds a vhost via rabbitmqctl add_vhost.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.add_vhost("saltstack") == {"Added": "saltstack"}


# 'delete_vhost' function tests: 1
def test_delete_vhost():
    """
    Test if it deletes a vhost rabbitmqctl delete_vhost.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.delete_vhost("saltstack") == {"Deleted": "saltstack"}


# 'set_permissions' function tests: 1
def test_set_permissions():
    """
    Test if it sets permissions for vhost via rabbitmqctl set_permissions.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.set_permissions("myvhost", "myuser") == {
            "Permissions Set": "saltstack"
        }


# 'list_permissions' function tests: 1
def test_list_permissions():
    """
    Test if it lists permissions for a vhost
    via rabbitmqctl list_permissions.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                '[{"user":"myuser","configure":"saltstack","write":".*","read":"1"}]'
            ),
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_permissions("saltstack") == {
            "myuser": {"configure": "saltstack", "write": ".*", "read": "1"},
        }


# 'list_user_permissions' function tests: 1
def test_list_user_permissions():
    """
    Test if it list permissions for a user
    via rabbitmqctl list_user_permissions.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": '[{"vhost":"saltstack","configure":"saltstack","write":"0","read":"1"},{"vhost":"guest","configure":"0","write":"one","read":""}]',
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_user_permissions("myuser") == {
            "saltstack": {"configure": "saltstack", "write": "0", "read": "1"},
            "guest": {"configure": "0", "write": "one", "read": ""},
        }


# 'set_user_tags' function tests: 1
def test_set_user_tags():
    """
    Test if it add user tags via rabbitmqctl set_user_tags.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.set_user_tags("myadmin", "admin") == {"Tag(s) set": "saltstack"}


# 'status' function tests: 1
def test_status():
    """
    Test if it return rabbitmq status.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.status() == "saltstack"


# 'cluster_status' function tests: 1
def test_cluster_status():
    """
    Test if it return rabbitmq cluster_status.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.cluster_status() == "saltstack"


# 'join_cluster' function tests: 1
def test_join_cluster():
    """
    Test if it join a rabbit cluster.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.join_cluster("rabbit.example.com") == {"Join": "saltstack"}


# 'stop_app' function tests: 1
def test_stop_app():
    """
    Test if it stops the RabbitMQ application,
    leaving the Erlang node running.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.stop_app() == "saltstack"


# 'start_app' function tests: 1
def test_start_app():
    """
    Test if it start the RabbitMQ application.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.start_app() == "saltstack"


# 'reset' function tests: 1
def test_reset():
    """
    Test if it return a RabbitMQ node to its virgin state
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.reset() == "saltstack"


# 'force_reset' function tests: 1
def test_force_reset():
    """
    Test if it forcefully Return a RabbitMQ node to its virgin state
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.force_reset() == "saltstack"


# 'list_queues' function tests: 1
def test_list_queues():
    """
    Test if it returns queue details of the / virtual host
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "saltstack\t0\nceleryev.234-234\t10",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_queues() == {
            "saltstack": ["0"],
            "celeryev.234-234": ["10"],
        }


# 'list_queues_vhost' function tests: 1
def test_list_queues_vhost():
    """
    Test if it returns queue details of specified virtual host.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": "saltstack\t0\nceleryev.234-234\t10",
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.list_queues_vhost("consumers") == {
            "saltstack": ["0"],
            "celeryev.234-234": ["10"],
        }


# 'list_policies' function tests: 3
def test_list_policies():
    """
    Test if it return a dictionary of policies nested by vhost
    and name based on the data returned from rabbitmqctl list_policies.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="3.7")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ), patch.dict(rabbitmq.__grains__, {"os_family": ""}):
        assert rabbitmq.list_policies() == {}


def test_list_policies_freebsd():
    """
    Test if it return a dictionary of policies nested by vhost
    and name based on the data returned from rabbitmqctl list_policies.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="3.7")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ), patch.dict(rabbitmq.__grains__, {"os_family": "FreeBSD"}):
        assert rabbitmq.list_policies() == {}


def test_list_policies_old_version():
    """
    Test if it return a dictionary of policies nested by vhost
    and name based on the data returned from rabbitmqctl list_policies.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="3.0")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ), patch.dict(rabbitmq.__grains__, {"os_family": ""}):
        assert rabbitmq.list_policies() == {}


# 'set_policy' function tests: 1
def test_set_policy():
    """
    Test if it set a policy based on rabbitmqctl set_policy.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.set_policy("/", "HA", ".*", '{"ha-mode": "all"}') == {
            "Set": "saltstack"
        }


# 'delete_policy' function tests: 1
def test_delete_policy():
    """
    Test if it delete a policy based on rabbitmqctl clear_policy.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.delete_policy("/", "HA") == {"Deleted": "saltstack"}


# 'policy_exists' function tests: 1
def test_policy_exists():
    """
    Test if it return whether the policy exists
    based on rabbitmqctl list_policies.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="3.0")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ), patch.dict(rabbitmq.__grains__, {"os_family": ""}):
        assert not rabbitmq.policy_exists("/", "HA")


# 'list_available_plugins' function tests: 2
def test_list_available_plugins():
    """
    Test if it returns a list of plugins.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack\nsalt\nother", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.list_available_plugins() == ["saltstack", "salt", "other"]


def test_list_available_plugins_space_delimited():
    """
    Test if it returns a list of plugins.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack salt other", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.list_available_plugins() == ["saltstack", "salt", "other"]


# 'list_enabled_plugins' function tests: 2
def test_list_enabled_plugins():
    """
    Test if it returns a list of plugins.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack\nsalt\nother", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.list_enabled_plugins() == ["saltstack", "salt", "other"]


def test_list_enabled_plugins_space_delimited():
    """
    Test if it returns a list of plugins.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack salt other", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.list_enabled_plugins() == ["saltstack", "salt", "other"]


# 'plugin_is_enabled' function tests: 2
def test_plugin_is_enabled():
    """
    Test if it returns true for an enabled plugin.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack\nsalt\nother", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.plugin_is_enabled("saltstack")
        assert rabbitmq.plugin_is_enabled("salt")
        assert rabbitmq.plugin_is_enabled("other")


def test_plugin_is_enabled_negative():
    """
    Test if it returns false for a disabled plugin.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack\nother", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert not rabbitmq.plugin_is_enabled("salt")
        assert not rabbitmq.plugin_is_enabled("stack")
        assert not rabbitmq.plugin_is_enabled("random")


# 'enable_plugin' function tests: 1
def test_enable_plugin():
    """
    Test if it enable a RabbitMQ plugin via the rabbitmq-plugins command.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.enable_plugin("salt") == {"Enabled": "saltstack"}


# 'disable_plugin' function tests: 1
def test_disable_plugin():
    """
    Test if it disable a RabbitMQ plugin via the rabbitmq-plugins command.
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stdout": "saltstack", "stderr": ""}
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.disable_plugin("salt") == {"Disabled": "saltstack"}


# 'list_upstreams' function tests: 1
def test_list_upstreams():
    """
    Test if it returns a list of upstreams.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                'federation-upstream\tremote-name\t{"ack-mode":"on-confirm"'
                ',"max-hops":1,"trust-user-id":true,"uri":"amqp://username:'
                'password@remote.fqdn"}'
            ),
            "stderr": "",
        }
    )
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        rabbitmq.__salt__, {"cmd.run_all": mock_run, "pkg.version": mock_pkg}
    ):
        assert rabbitmq.list_upstreams() == {
            "remote-name": (
                '{"ack-mode":"on-confirm","max-hops":1,'
                '"trust-user-id":true,"uri":"amqp://username:'
                'password@remote.fqdn"}'
            )
        }


# 'upstream_exists' function tests: 2
def test_upstream_exists():
    """
    Test whether a given rabbitmq-internal upstream exists based
    on rabbitmqctl list_upstream.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                'federation-upstream\tremote-name\t{"ack-mode":"on-confirm"'
                ',"max-hops":1,"trust-user-id":true,"uri":"amqp://username:'
                'password@remote.fqdn"}'
            ),
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.upstream_exists("remote-name")


def test_upstream_exists_negative():
    """
    Negative test of whether rabbitmq-internal upstream exists based
    on rabbitmqctl list_upstream.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                'federation-upstream\tremote-name\t{"ack-mode":"on-confirm"'
                ',"max-hops":1,"trust-user-id":true,"uri":"amqp://username:'
                'password@remote.fqdn"}'
            ),
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert not rabbitmq.upstream_exists("does-not-exist")


# 'add_upstream' function tests: 1
def test_set_upstream():
    """
    Test if a rabbitMQ upstream gets configured properly.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                'Setting runtime parameter "federation-upstream" for component '
                '"remote-name" to "{"trust-user-id": true, "uri": '
                '"amqp://username:password@remote.fqdn", "ack-mode": "on-confirm", '
                '"max-hops": 1}" in vhost "/" ...'
            ),
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.set_upstream(
            "remote-name",
            "amqp://username:password@remote.fqdn",
            ack_mode="on-confirm",
            max_hops=1,
            trust_user_id=True,
        )


# 'delete_upstream' function tests: 2
def test_delete_upstream():
    """
    Test if an upstream gets deleted properly using rabbitmqctl delete_upstream.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": (
                'Clearing runtime parameter "remote-name" for component '
                '"federation-upstream" on vhost "/" ...'
            ),
            "stderr": "",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        assert rabbitmq.delete_upstream("remote-name")


def test_delete_upstream_negative():
    """
    Negative test trying to delete a non-existant upstream.
    """
    mock_run = MagicMock(
        return_value={
            "retcode": 70,
            "stdout": (
                'Clearing runtime parameter "remote-name" for component '
                '"federation-upstream" on vhost "/" ...'
            ),
            "stderr": "Error:\nParameter does not exist",
        }
    )
    with patch.dict(rabbitmq.__salt__, {"cmd.run_all": mock_run}):
        pytest.raises(CommandExecutionError, rabbitmq.delete_upstream, "remote-name")



"""
Parses roster entries out of Host directives from SSH known_hosts

.. versionadded:: 3006.0

Sample configuration:

.. note::

    The ``known_hosts`` file only contains hostname/IP. To pass other parameters,
    use ``roster_defaults``.

.. code-block:: yaml

    ssh_known_hosts_file: /Users/user1/.ssh/known_hosts
    roster_defaults:
      user: user1
      sudo: True

Now you can use the module

.. code-block:: bash

    salt-ssh --roster sshknownhosts '*' -r "echo hi"

Or with a Saltfile

.. code-block:: yaml

    salt-ssh:
      ssh_known_hosts_file: /Users/user1/.ssh/known_hosts

.. code-block:: bash

    salt-ssh --roster sshknownhosts '*' -r "echo hi"

"""


import logging
import os

import salt.utils.files
import salt.utils.stringutils

log = logging.getLogger(__name__)


def _parse_ssh_known_hosts_line(line):
    """
    Parse one line from a known_hosts line

    :param line: Individual lines from the ssh known_hosts file
    :return: Dict that contain the three fields from a known_hosts line
    """
    line_unicode = salt.utils.stringutils.to_unicode(line)
    fields = line_unicode.split(" ")

    if len(fields) < 3:
        log.warn("Not enough fields found in known_hosts in line : %s", line)
        return None

    fields = fields[:3]

    names, keytype, key = fields
    names = names.split(",")

    return {"names": names, "keytype": keytype, "key": key}


def _parse_ssh_known_hosts(lines):
    """
    Parses lines from the SSH known_hosts to create roster targets.

    :param lines: lines from the ssh known_hosts file
    :return: Dictionary of targets in similar style to the flat roster
    """

    targets_ = {}
    for line in lines:
        host_key = _parse_ssh_known_hosts_line(line)

        for host in host_key["names"]:
            targets_.update({host: {"host": host}})

    return targets_


def targets(tgt, tgt_type="glob"):
    """
    Return the targets from a known_hosts file
    """

    ssh_known_hosts_file = __opts__.get("ssh_known_hosts_file")

    if not os.path.isfile(ssh_known_hosts_file):
        log.error("Cannot find SSH known_hosts file")
        raise OSError("Cannot find SSH known_hosts file")
    if not os.access(ssh_known_hosts_file, os.R_OK):
        log.error("Cannot access SSH known_hosts file: %s", ssh_known_hosts_file)
        raise OSError(
            "Cannot access SSH known_hosts file: {}".format(ssh_known_hosts_file)
        )

    with salt.utils.files.fopen(ssh_known_hosts_file, "r") as hostfile:
        raw = _parse_ssh_known_hosts([line.rstrip() for line in hostfile])

    return __utils__["roster_matcher.targets"](raw, tgt, tgt_type, "ipv4")



"""
Management of Portage package configuration on Gentoo
=====================================================

A state module to manage Portage configuration on Gentoo

.. code-block:: yaml

    salt:
        portage_config.flags:
            - use:
                - openssl
"""


def __virtual__():
    """
    Only load if the portage_config module is available in __salt__
    """
    if "portage_config.get_missing_flags" in __salt__:
        return "portage_config"
    return (False, "portage module could not be loaded")


def mod_init(low):
    """
    Enforce a nice structure on the configuration files.
    """
    try:
        __salt__["portage_config.enforce_nice_config"]()
    except Exception:  # pylint: disable=broad-except
        return False
    return True


def _flags_helper(conf, atom, new_flags, test=False):
    try:
        new_flags = __salt__["portage_config.get_missing_flags"](conf, atom, new_flags)
    except Exception:  # pylint: disable=broad-except
        import traceback

        return {"result": False, "comment": traceback.format_exc()}
    if new_flags:
        old_flags = __salt__["portage_config.get_flags_from_package_conf"](conf, atom)
        if not test:
            __salt__["portage_config.append_to_package_conf"](conf, atom, new_flags)
        return {"result": True, "changes": {"old": old_flags, "new": new_flags}}
    return {"result": None}


def _mask_helper(conf, atom, test=False):
    try:
        is_present = __salt__["portage_config.is_present"](conf, atom)
    except Exception:  # pylint: disable=broad-except
        import traceback

        return {"result": False, "comment": traceback.format_exc()}
    if not is_present:
        if not test:
            __salt__["portage_config.append_to_package_conf"](conf, string=atom)
        return {"result": True}
    return {"result": None}


def flags(
    name,
    use=None,
    accept_keywords=None,
    env=None,
    license=None,
    properties=None,
    unmask=False,
    mask=False,
):
    """
    Enforce the given flags on the given package or ``DEPEND`` atom.

    .. warning::

        In most cases, the affected package(s) need to be rebuilt in
        order to apply the changes.

    name
        The name of the package or its DEPEND atom
    use
        A list of ``USE`` flags
    accept_keywords
        A list of keywords to accept. ``~ARCH`` means current host arch, and will
        be translated into a line without keywords
    env
        A list of environment files
    license
        A list of accepted licenses
    properties
        A list of additional properties
    unmask
        A boolean to unmask the package
    mask
        A boolean to mask the package
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    if use:
        result = _flags_helper("use", name, use, __opts__["test"])
        if result["result"]:
            ret["changes"]["use"] = result["changes"]
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if accept_keywords:
        result = _flags_helper(
            "accept_keywords", name, accept_keywords, __opts__["test"]
        )
        if result["result"]:
            ret["changes"]["accept_keywords"] = result["changes"]
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if env:
        result = _flags_helper("env", name, env, __opts__["test"])
        if result["result"]:
            ret["changes"]["env"] = result["changes"]
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if license:
        result = _flags_helper("license", name, license, __opts__["test"])
        if result["result"]:
            ret["changes"]["license"] = result["changes"]
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if properties:
        result = _flags_helper("properties", name, properties, __opts__["test"])
        if result["result"]:
            ret["changes"]["properties"] = result["changes"]
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if mask:
        result = _mask_helper("mask", name, __opts__["test"])
        if result["result"]:
            ret["changes"]["mask"] = "masked"
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if unmask:
        result = _mask_helper("unmask", name, __opts__["test"])
        if result["result"]:
            ret["changes"]["unmask"] = "unmasked"
        elif result["result"] is False:
            ret["result"] = False
            ret["comment"] = result["comment"]
            return ret

    if __opts__["test"] and not ret["result"]:
        ret["result"] = None

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
    :codeauthor: Andrew Colin Kissa <andrew@topdog.za.net>
"""

import os

import pytest

import salt.states.augeas as augeas
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {augeas: {}}


@pytest.fixture
def setup_func():
    name = "zabbix"
    context = "/files/etc/services"
    changes = [
        "ins service-name after service-name[last()]",
        "set service-name[last()] zabbix-agent",
    ]
    fp_changes = [
        "ins service-name after /files/etc/services/service-name[last()]",
        "set /files/etc/services/service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    method_map = {
        "set": "set",
        "setm": "setm",
        "mv": "move",
        "move": "move",
        "ins": "insert",
        "insert": "insert",
        "rm": "remove",
        "remove": "remove",
    }
    mock_method_map = MagicMock(return_value=method_map)
    return mock_method_map


def test_change_non_list_changes():
    """
    Test if none list changes handled correctly
    """
    name = "zabbix"
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    comt = "'changes' must be specified as a list"
    ret.update({"comment": comt})

    assert augeas.change(name) == ret


def test_change_non_list_load_path():
    """
    Test if none list load_path is handled correctly
    """
    name = "zabbix"
    context = "/files/etc/services"
    changes = [
        "ins service-name after service-name[last()]",
        "set service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    comt = "'load_path' must be specified as a list"
    ret.update({"comment": comt})

    assert augeas.change(name, context, changes, load_path="x") == ret


def test_change_in_test_mode():
    """
    Test test mode handling
    """
    name = "zabbix"
    context = "/files/etc/services"
    changes = [
        "ins service-name after service-name[last()]",
        "set service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    comt = (
        'Executing commands in file "/files/etc/services":\n'
        "ins service-name after service-name[last()]"
        "\nset service-name[last()] zabbix-agent"
    )
    ret.update({"comment": comt, "result": True})

    with patch.dict(augeas.__opts__, {"test": True}):
        assert augeas.change(name, context, changes) == ret


def test_change_no_context_without_full_path(setup_func):
    """
    Test handling of no context without full path
    """
    name = "zabbix"
    context = "/files/etc/services"
    changes = [
        "ins service-name after service-name[last()]",
        "set service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    comt = (
        "Error: Changes should be prefixed with /files if no "
        "context is provided, change: {}".format(changes[0])
    )
    ret.update({"comment": comt, "result": False})

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_dict_ = {"augeas.method_map": setup_func.mock_method_map}
        with patch.dict(augeas.__salt__, mock_dict_):
            assert augeas.change(name, changes=changes) == ret


def test_change_no_context_with_full_path_fail(setup_func):
    """
    Test handling of no context with full path with execute fail
    """
    name = "zabbix"
    fp_changes = [
        "ins service-name after /files/etc/services/service-name[last()]",
        "set /files/etc/services/service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    ret.update({"comment": "Error: error", "result": False})

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=False, error="error"))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            assert augeas.change(name, changes=fp_changes) == ret


def test_change_no_context_with_full_path_pass(setup_func):
    """
    Test handling of no context with full path with execute pass
    """
    name = "zabbix"
    fp_changes = [
        "ins service-name after /files/etc/services/service-name[last()]",
        "set /files/etc/services/service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    ret.update(
        dict(
            comment="Changes have been saved",
            result=True,
            changes={"diff": "+ zabbix-agent"},
        )
    )

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=True))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            mock_filename = MagicMock(return_value="/etc/services")
            with patch.object(augeas, "_workout_filename", mock_filename), patch(
                "os.path.isfile", MagicMock(return_value=True)
            ):
                with patch("salt.utils.files.fopen", MagicMock(mock_open)):
                    mock_diff = MagicMock(return_value=["+ zabbix-agent"])
                    with patch("difflib.unified_diff", mock_diff):
                        assert augeas.change(name, changes=fp_changes) == ret


def test_change_no_context_without_full_path_invalid_cmd(setup_func):
    """
    Test handling of invalid commands when no context supplied
    """
    name = "zabbix"
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    ret.update(dict(comment="Error: Command det is not supported (yet)", result=False))

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=True))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            changes = ["det service-name[last()] zabbix-agent"]
            assert augeas.change(name, changes=changes) == ret


def test_change_no_context_without_full_path_invalid_change(setup_func):
    """
    Test handling of invalid change when no context supplied
    """
    name = "zabbix"
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    comt = "Error: Invalid formatted command, see debug log for details: require"
    ret.update(dict(comment=comt, result=False))
    changes = ["require"]

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=True))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            assert augeas.change(name, changes=changes) == ret


def test_change_no_context_with_full_path_multiple_files(setup_func):
    """
    Test handling of different paths with no context supplied
    """
    name = "zabbix"
    changes = [
        "set /files/etc/hosts/service-name test",
        "set /files/etc/services/service-name test",
    ]
    filename = "/etc/hosts/service-name"
    filename_ = "/etc/services/service-name"
    comt = (
        "Error: Changes should be made to one file at a time, "
        "detected changes to {} and {}".format(filename, filename_)
    )
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    ret.update(dict(comment=comt, result=False))

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=True))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            assert augeas.change(name, changes=changes) == ret


def test_change_with_context_without_full_path_fail(setup_func):
    """
    Test handling of context without full path fails
    """
    name = "zabbix"
    context = "/files/etc/services"
    changes = [
        "ins service-name after service-name[last()]",
        "set service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    ret.update(dict(comment="Error: error", result=False))

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=False, error="error"))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            with patch("salt.utils.files.fopen", MagicMock(mock_open)):
                assert augeas.change(name, context=context, changes=changes) == ret


def test_change_with_context_without_old_file(setup_func):
    """
    Test handling of context without oldfile pass
    """
    name = "zabbix"
    context = "/files/etc/services"
    changes = [
        "ins service-name after service-name[last()]",
        "set service-name[last()] zabbix-agent",
    ]
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    ret.update(
        dict(
            comment="Changes have been saved",
            result=True,
            changes={"updates": changes},
        )
    )

    with patch.dict(augeas.__opts__, {"test": False}):
        mock_execute = MagicMock(return_value=dict(retval=True))
        mock_dict_ = {
            "augeas.execute": mock_execute,
            "augeas.method_map": setup_func.mock_method_map,
        }
        with patch.dict(augeas.__salt__, mock_dict_):
            mock_isfile = MagicMock(return_value=False)
            with patch.object(os.path, "isfile", mock_isfile):
                assert augeas.change(name, context=context, changes=changes) == ret



import pytest

import salt.modules.mac_softwareupdate as mac_softwareupdate
from tests.support.mock import MagicMock, patch

MOJAVE_LIST_OUTPUT = """Software Update Tool

Finding available software
Software Update found the following new or updated software:
   * Command Line Tools (macOS Mojave version 10.14) for Xcode-10.3
    Command Line Tools (macOS Mojave version 10.14) for Xcode (10.3), 199140K [recommended]
   * macOS 10.14.1 Update
    macOS 10.14.1 Update (10.14.1), 199140K [recommended] [restart]
   * BridgeOSUpdateCustomer
    BridgeOSUpdateCustomer (10.14.4.1.1.1555388607), 328394K, [recommended] [shut down]
   - iCal-1.0.2
    iCal, (1.0.2), 6520K"""


CATALINA_LIST_OUTPUT = """Software Update Tool

Finding available software
Software Update found the following new or updated software:
* Label: Command Line Tools beta 5 for Xcode-11.0
    Title: Command Line Tools beta 5 for Xcode, Version: 11.0, Size: 224804K, Recommended: YES,
* Label: macOS Catalina Developer Beta-6
    Title: macOS Catalina Public Beta, Version: 5, Size: 3084292K, Recommended: YES, Action: restart,
* Label: BridgeOSUpdateCustomer
    Title: BridgeOSUpdateCustomer, Version: 10.15.0.1.1.1560926689, Size: 390674K, Recommended: YES, Action: shut down,
- Label: iCal-1.0.2
    Title: iCal, Version: 1.0.2, Size: 6520K,"""


@pytest.fixture
def configure_loader_modules():
    return {mac_softwareupdate: {}}


def test_mojave_list_available():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 14, 6]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=MOJAVE_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available()
            expected = {
                "Command Line Tools (macOS Mojave version 10.14) for Xcode-10.3": "10.3",
                "macOS 10.14.1 Update": "10.14.1",
                "BridgeOSUpdateCustomer": "10.14.4.1.1.1555388607",
                "iCal-1.0.2": "1.0.2",
            }
            assert result == expected


def test_mojave_list_available_trailing_ws():
    """Ensure the regex works with trailing whitespace in labels"""
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 14, 6]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(
                return_value=(
                    "Software Update Tool\n\nFinding available software\nSoftware Update found"
                    " the following new or updated software:\n   * macOS Mojave 10.14.6"
                    " Supplemental Update- \n    macOS Mojave 10.14.6 Supplemental Update ( ),"
                    " 1581834K [recommended] [restart]"
                )
            ),
        ):
            result = mac_softwareupdate.list_available()
            expected = {"macOS Mojave 10.14.6 Supplemental Update- ": ""}
            assert result == expected


def test_mojave_list_recommended():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 14, 6]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=MOJAVE_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(recommended=True)
            expected = {
                "Command Line Tools (macOS Mojave version 10.14) for Xcode-10.3": "10.3",
                "macOS 10.14.1 Update": "10.14.1",
                "BridgeOSUpdateCustomer": "10.14.4.1.1.1555388607",
            }
            assert result == expected


def test_mojave_list_restart():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 14, 6]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=MOJAVE_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(restart=True)
            expected = {"macOS 10.14.1 Update": "10.14.1"}
            assert result == expected


def test_mojave_list_shut_down():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 14, 6]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=MOJAVE_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(shut_down=True)
            expected = {"BridgeOSUpdateCustomer": "10.14.4.1.1.1555388607"}
            assert result == expected


def test_catalina_list_available():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 15]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available()
            expected = {
                "Command Line Tools beta 5 for Xcode-11.0": "11.0",
                "macOS Catalina Developer Beta-6": "5",
                "BridgeOSUpdateCustomer": "10.15.0.1.1.1560926689",
                "iCal-1.0.2": "1.0.2",
            }
            assert result == expected


def test_catalina_list_recommended():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 15]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(recommended=True)
            expected = {
                "Command Line Tools beta 5 for Xcode-11.0": "11.0",
                "macOS Catalina Developer Beta-6": "5",
                "BridgeOSUpdateCustomer": "10.15.0.1.1.1560926689",
            }
            assert result == expected


def test_catalina_list_restart():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 15]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(restart=True)
            expected = {"macOS Catalina Developer Beta-6": "5"}
            assert result == expected


def test_catalina_list_shut_down():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [10, 15]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(shut_down=True)
            expected = {"BridgeOSUpdateCustomer": "10.15.0.1.1.1560926689"}
            assert result == expected


def test_bigsur_list_available():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [11, 0]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available()
            expected = {
                "Command Line Tools beta 5 for Xcode-11.0": "11.0",
                "macOS Catalina Developer Beta-6": "5",
                "BridgeOSUpdateCustomer": "10.15.0.1.1.1560926689",
                "iCal-1.0.2": "1.0.2",
            }
            assert result == expected


def test_bigsur_list_recommended():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [11, 0]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(recommended=True)
            expected = {
                "Command Line Tools beta 5 for Xcode-11.0": "11.0",
                "macOS Catalina Developer Beta-6": "5",
                "BridgeOSUpdateCustomer": "10.15.0.1.1.1560926689",
            }
            assert result == expected


def test_bigsur_list_restart():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [11, 0]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(restart=True)
            expected = {"macOS Catalina Developer Beta-6": "5"}
            assert result == expected


def test_bigsur_list_shut_down():
    with patch.dict(mac_softwareupdate.__grains__, {"osrelease_info": [11, 0]}):
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value=CATALINA_LIST_OUTPUT),
        ):
            result = mac_softwareupdate.list_available(shut_down=True)
            expected = {"BridgeOSUpdateCustomer": "10.15.0.1.1.1560926689"}
            assert result == expected



"""
Module for interfacing with SysFS

.. seealso:: https://www.kernel.org/doc/Documentation/filesystems/sysfs.txt
.. versionadded:: 2016.3.0
"""

import logging
import os
import stat

import salt.utils.files
import salt.utils.path
import salt.utils.platform

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work on Linux
    """
    return salt.utils.platform.is_linux()


def attr(key, value=None):
    """
    Access/write a SysFS attribute.
    If the attribute is a symlink, its destination is returned

    :return: value or bool

    CLI Example:

     .. code-block:: bash

        salt '*' sysfs.attr block/sda/queue/logical_block_size
    """
    key = target(key)

    if key is False:
        return False
    elif os.path.isdir(key):
        return key
    elif value is not None:
        return write(key, value)
    else:
        return read(key)


def write(key, value):
    """
    Write a SysFS attribute/action

    CLI Example:

     .. code-block:: bash

        salt '*' sysfs.write devices/system/cpu/cpu0/cpufreq/scaling_governor 'performance'
    """
    try:
        key = target(key)
        log.trace("Writing %s to %s", value, key)
        with salt.utils.files.fopen(key, "w") as twriter:
            twriter.write(salt.utils.stringutils.to_str("{}\n".format(value)))
            return True
    except Exception:  # pylint: disable=broad-except
        return False


def read(key, root=""):
    """
    Read from SysFS

    :param key: file or path in SysFS; if key is a list then root will be prefixed on each key

    :return: the full (tree of) SysFS attributes under key

    CLI Example:

     .. code-block:: bash

        salt '*' sysfs.read class/net/em1/statistics
    """

    if not isinstance(key, str):
        res = {}
        for akey in key:
            ares = read(os.path.join(root, akey))
            if ares is not False:
                res[akey] = ares
        return res

    key = target(os.path.join(root, key))
    if key is False:
        return False
    elif os.path.isdir(key):
        keys = interfaces(key)
        result = {}
        for subkey in keys["r"] + keys["rw"]:
            subval = read(os.path.join(key, subkey))
            if subval is not False:
                subkeys = subkey.split("/")
                subkey = subkeys.pop()
                subresult = result
                if subkeys:
                    for skey in subkeys:
                        if skey not in subresult:
                            subresult[skey] = {}
                        subresult = subresult[skey]
                subresult[subkey] = subval
        return result
    else:
        try:
            log.trace("Reading %s...", key)

            # Certain things in SysFS are pipes 'n such.
            # This opens it non-blocking, which prevents indefinite blocking
            with os.fdopen(os.open(key, os.O_RDONLY | os.O_NONBLOCK)) as treader:
                # alternative method for the same idea, but only works for completely empty pipes
                # treader = select.select([treader], [], [], 1)[0][0]
                val = treader.read().strip()
                if not val:
                    return False
                try:
                    val = int(val)
                except Exception:  # pylint: disable=broad-except
                    try:
                        val = float(val)
                    except Exception:  # pylint: disable=broad-except
                        pass
                return val
        except Exception:  # pylint: disable=broad-except
            return False


def target(key, full=True):
    """
    Return the basename of a SysFS key path

    :param key: the location to resolve within SysFS
    :param full: full path instead of basename

    :return: fullpath or basename of path

    CLI Example:

     .. code-block:: bash

        salt '*' sysfs.read class/ttyS0

    """
    if not key.startswith("/sys"):
        key = os.path.join("/sys", key)
    key = os.path.realpath(key)

    if not os.path.exists(key):
        log.debug("Unknown SysFS key %s", key)
        return False
    elif full:
        return key
    else:
        return os.path.basename(key)


def interfaces(root):
    """
    Generate a dictionary with all available interfaces relative to root.
    Symlinks are not followed.

    CLI Example:

     .. code-block:: bash

        salt '*' sysfs.interfaces block/bcache0/bcache

    Output example:
     .. code-block:: json

       {
          "r": [
            "state",
            "partial_stripes_expensive",
            "writeback_rate_debug",
            "stripe_size",
            "dirty_data",
            "stats_total/cache_hits",
            "stats_total/cache_bypass_misses",
            "stats_total/bypassed",
            "stats_total/cache_readaheads",
            "stats_total/cache_hit_ratio",
            "stats_total/cache_miss_collisions",
            "stats_total/cache_misses",
            "stats_total/cache_bypass_hits",
          ],
          "rw": [
            "writeback_rate",
            "writeback_rate_update_seconds",
            "cache_mode",
            "writeback_delay",
            "label",
            "writeback_running",
            "writeback_metadata",
            "running",
            "writeback_rate_p_term_inverse",
            "sequential_cutoff",
            "writeback_percent",
            "writeback_rate_d_term",
            "readahead"
          ],
          "w": [
            "stop",
            "clear_stats",
            "attach",
            "detach"
          ]
       }

    .. note::
      * 'r' interfaces are read-only
      * 'w' interfaces are write-only (e.g. actions)
      * 'rw' are interfaces that can both be read or written
    """

    root = target(root)
    if root is False or not os.path.isdir(root):
        log.error("SysFS %s not a dir", root)
        return False

    readwrites = []
    reads = []
    writes = []

    for path, _, files in salt.utils.path.os_walk(root, followlinks=False):
        for afile in files:
            canpath = os.path.join(path, afile)

            if not os.path.isfile(canpath):
                continue

            stat_mode = os.stat(canpath).st_mode
            is_r = bool(stat.S_IRUSR & stat_mode)
            is_w = bool(stat.S_IWUSR & stat_mode)

            relpath = os.path.relpath(canpath, root)
            if is_w:
                if is_r:
                    readwrites.append(relpath)
                else:
                    writes.append(relpath)
            elif is_r:
                reads.append(relpath)
            else:
                log.warning("Unable to find any interfaces in %s", canpath)

    return {"r": reads, "w": writes, "rw": readwrites}



"""
Opening of Encrypted Devices
============================

Ensure that an encrypted device is mapped with the `mapped` function:

.. code-block:: yaml

    mappedname:
      cryptdev.mapped:
        - device: /dev/sdb1
        - keyfile: /etc/keyfile.key
        - opts:
          - size=256

    swap:
      crypted.mapped:
        - device: /dev/sdx4
        - keyfile: /dev/urandom
        - opts: swap,cipher=aes-cbc-essiv:sha256,size=256

    mappedbyuuid:
      crypted.mapped:
        - device: UUID=066e0200-2867-4ebe-b9e6-f30026ca2314
        - keyfile: /etc/keyfile.key
        - config: /etc/alternate-crypttab

.. versionadded:: 2018.3.0
"""

import logging

log = logging.getLogger(__name__)


def mapped(
    name,
    device,
    keyfile=None,
    opts=None,
    config="/etc/crypttab",
    persist=True,
    immediate=False,
    match_on="name",
):
    """
    Verify that a device is mapped

    name
        The name under which the device is to be mapped

    device
        The device name, typically the device node, such as ``/dev/sdb1``
        or ``UUID=066e0200-2867-4ebe-b9e6-f30026ca2314``.

    keyfile
        Either ``None`` if the password is to be entered manually on boot, or
        an absolute path to a keyfile. If the password is to be asked
        interactively, the mapping cannot be performed with ``immediate=True``.

    opts
        A list object of options or a comma delimited list

    config
        Set an alternative location for the crypttab, if the map is persistent,
        Default is ``/etc/crypttab``

    persist
        Set if the map should be saved in the crypttab, Default is ``True``

    immediate
        Set if the device mapping should be executed immediately. Requires that
        the keyfile not be ``None``, because the password cannot be asked
        interactively. Note that options are not passed through on the initial
        mapping.  Default is ``False``.

    match_on
        A name or list of crypttab properties on which this state should be applied.
        Default is ``name``, meaning that the line is matched only by the name
        parameter. If the desired configuration requires two devices mapped to
        the same name, supply a list of parameters to match on.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # If neither option is set, we've been asked to do nothing.
    if not immediate and not persist:
        ret["result"] = False
        ret[
            "comment"
        ] = "Either persist or immediate must be set, otherwise this state does nothing"
        return ret

    if immediate and (keyfile is None or keyfile == "none" or keyfile == "-"):
        ret["result"] = False
        ret["changes"][
            "cryptsetup"
        ] = "Device cannot be mapped immediately without a keyfile"

    elif immediate:
        # Get the active crypt mounts. If ours is listed already, no action is necessary.
        active = __salt__["cryptdev.active"]()
        if name not in active.keys():
            # Open the map using cryptsetup. This does not pass any options.
            if opts:
                log.warning("Ignore cryptdev configuration when mapping immediately")

            if __opts__["test"]:
                ret["result"] = None
                ret["commment"] = "Device would be mapped immediately"
            else:
                cryptsetup_result = __salt__["cryptdev.open"](name, device, keyfile)
                if cryptsetup_result:
                    ret["changes"]["cryptsetup"] = "Device mapped using cryptsetup"
                else:
                    ret["changes"][
                        "cryptsetup"
                    ] = "Device failed to map using cryptsetup"
                    ret["result"] = False

    if persist and not __opts__["test"]:
        crypttab_result = __salt__["cryptdev.set_crypttab"](
            name,
            device,
            password=keyfile,
            options=opts,
            config=config,
            match_on=match_on,
        )
        if crypttab_result:
            if crypttab_result == "new":
                ret["changes"]["crypttab"] = "Entry added in {}".format(config)

            if crypttab_result == "change":
                ret["changes"]["crypttab"] = "Existing entry in {} changed".format(
                    config
                )

        else:
            ret["changes"]["crypttab"] = "Unable to set entry in {}".format(config)
            ret["result"] = False

    return ret


def unmapped(name, config="/etc/crypttab", persist=True, immediate=False):
    """
    Ensure that a device is unmapped

    name
        The name to ensure is not mapped

    config
        Set an alternative location for the crypttab, if the map is persistent,
        Default is ``/etc/crypttab``

    persist
        Set if the map should be removed from the crypttab. Default is ``True``

    immediate
        Set if the device should be unmapped immediately. Default is ``False``.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if immediate:
        # Get the active crypt mounts. If ours is not listed already, no action is necessary.
        active = __salt__["cryptdev.active"]()
        if name in active.keys():
            # Close the map using cryptsetup.
            if __opts__["test"]:
                ret["result"] = None
                ret["commment"] = "Device would be unmapped immediately"
            else:
                cryptsetup_result = __salt__["cryptdev.close"](name)
                if cryptsetup_result:
                    ret["changes"]["cryptsetup"] = "Device unmapped using cryptsetup"
                else:
                    ret["changes"][
                        "cryptsetup"
                    ] = "Device failed to unmap using cryptsetup"
                    ret["result"] = False

    if persist and not __opts__["test"]:
        crypttab_result = __salt__["cryptdev.rm_crypttab"](name, config=config)
        if crypttab_result:
            if crypttab_result == "change":
                ret["changes"]["crypttab"] = "Entry removed from {}".format(config)

        else:
            ret["changes"]["crypttab"] = "Unable to remove entry in {}".format(config)
            ret["result"] = False

    return ret



"""
Support for eselect, Gentoo's configuration and management tool.
"""


import logging

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work on Gentoo systems with eselect installed
    """
    if __grains__["os"] == "Gentoo" and salt.utils.path.which("eselect"):
        return "eselect"
    return (
        False,
        "The eselect execution module cannot be loaded: either the system is not Gentoo"
        " or the eselect binary is not in the path.",
    )


def exec_action(
    module, action, module_parameter=None, action_parameter=None, state_only=False
):
    """
    Execute an arbitrary action on a module.

    module
        name of the module to be executed

    action
        name of the module's action to be run

    module_parameter
        additional params passed to the defined module

    action_parameter
        additional params passed to the defined action

    state_only
        don't return any output but only the success/failure of the operation

    CLI Example (updating the ``php`` implementation used for ``apache2``):

    .. code-block:: bash

        salt '*' eselect.exec_action php update action_parameter='apache2'
    """
    out = __salt__["cmd.run"](
        "eselect --brief --colour=no {} {} {} {}".format(
            module, module_parameter or "", action, action_parameter or ""
        ),
        python_shell=False,
    )
    out = out.strip().split("\n")

    if out[0].startswith("!!! Error"):
        return False

    if state_only:
        return True

    if not out:
        return False

    if len(out) == 1 and not out[0].strip():
        return False

    return out


def get_modules():
    """
    List available ``eselect`` modules.

    CLI Example:

    .. code-block:: bash

        salt '*' eselect.get_modules
    """
    modules = []
    module_list = exec_action("modules", "list", action_parameter="--only-names")
    if not module_list:
        return None

    for module in module_list:
        if module not in ["help", "usage", "version"]:
            modules.append(module)
    return modules


def get_target_list(module, action_parameter=None):
    """
    List available targets for the given module.

    module
        name of the module to be queried for its targets

    action_parameter
        additional params passed to the defined action

        .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt '*' eselect.get_target_list kernel
    """
    exec_output = exec_action(module, "list", action_parameter=action_parameter)
    if not exec_output:
        return None

    target_list = []
    if isinstance(exec_output, list):
        for item in exec_output:
            target_list.append(item.split(None, 1)[0])
        return target_list

    return None


def get_current_target(module, module_parameter=None, action_parameter=None):
    """
    Get the currently selected target for the given module.

    module
        name of the module to be queried for its current target

    module_parameter
        additional params passed to the defined module

    action_parameter
        additional params passed to the 'show' action

    CLI Example (current target of system-wide ``java-vm``):

    .. code-block:: bash

        salt '*' eselect.get_current_target java-vm action_parameter='system'

    CLI Example (current target of ``kernel`` symlink):

    .. code-block:: bash

        salt '*' eselect.get_current_target kernel
    """
    result = exec_action(
        module,
        "show",
        module_parameter=module_parameter,
        action_parameter=action_parameter,
    )[0]
    if not result:
        return None

    if result == "(unset)":
        return None

    return result


def set_target(module, target, module_parameter=None, action_parameter=None):
    """
    Set the target for the given module.
    Target can be specified by index or name.

    module
        name of the module for which a target should be set

    target
        name of the target to be set for this module

    module_parameter
        additional params passed to the defined module

    action_parameter
        additional params passed to the defined action

    CLI Example (setting target of system-wide ``java-vm``):

    .. code-block:: bash

        salt '*' eselect.set_target java-vm icedtea-bin-7 action_parameter='system'

    CLI Example (setting target of ``kernel`` symlink):

    .. code-block:: bash

        salt '*' eselect.set_target kernel linux-3.17.5-gentoo
    """
    if action_parameter:
        action_parameter = "{} {}".format(action_parameter, target)
    else:
        action_parameter = target

    # get list of available modules
    if module not in get_modules():
        log.error("Module %s not available", module)
        return False

    exec_result = exec_action(
        module,
        "set",
        module_parameter=module_parameter,
        action_parameter=action_parameter,
        state_only=True,
    )
    if exec_result:
        return exec_result
    return False



import pytest

from tests.support.case import SyndicCase


@pytest.mark.windows_whitelisted
@pytest.mark.skip(reason="The Syndic Tests are currently broken. See #58975")
class TestSyndic(SyndicCase):
    """
    Validate the syndic interface by testing the test module
    """

    @pytest.mark.slow_test
    def test_ping(self):
        """
        test.ping
        """
        self.assertTrue(self.run_function("test.ping"))

    @pytest.mark.slow_test
    def test_fib(self):
        """
        test.fib
        """
        self.assertEqual(
            self.run_function(
                "test.fib",
                ["20"],
            )[0],
            6765,
        )



"""
Test the RSA ANSI X9.31 signer and verifier
"""

import ctypes
import ctypes.util
import fnmatch
import glob
import os
import platform
import sys

import pytest

import salt.utils.platform
from salt.utils.rsax931 import (
    RSAX931Signer,
    RSAX931Verifier,
    _find_libcrypto,
    _load_libcrypto,
)
from tests.support.mock import patch


@pytest.fixture
def privkey_data():
    return (
        "-----BEGIN RSA PRIVATE KEY-----\n"
        "MIIEpAIBAAKCAQEA75GR6ZTv5JOv90Vq8tKhKC7YQnhDIo2hM0HVziTEk5R4UQBW\n"
        "a0CKytFMbTONY2msEDwX9iA0x7F5Lgj0X8eD4ZMsYqLzqjWMekLC8bjhxc+EuPo9\n"
        "Dygu3mJ2VgRC7XhlFpmdo5NN8J2E7B/CNB3R4hOcMMZNZdi0xLtFoTfwU61UPfFX\n"
        "14mV2laqLbvDEfQLJhUTDeFFV8EN5Z4H1ttLP3sMXJvc3EvM0JiDVj4l1TWFUHHz\n"
        "eFgCA1Im0lv8i7PFrgW7nyMfK9uDSsUmIp7k6ai4tVzwkTmV5PsriP1ju88Lo3MB\n"
        "4/sUmDv/JmlZ9YyzTO3Po8Uz3Aeq9HJWyBWHAQIDAQABAoIBAGOzBzBYZUWRGOgl\n"
        "IY8QjTT12dY/ymC05GM6gMobjxuD7FZ5d32HDLu/QrknfS3kKlFPUQGDAbQhbbb0\n"
        "zw6VL5NO9mfOPO2W/3FaG1sRgBQcerWonoSSSn8OJwVBHMFLG3a+U1Zh1UvPoiPK\n"
        "S734swIM+zFpNYivGPvOm/muF/waFf8tF/47t1cwt/JGXYQnkG/P7z0vp47Irpsb\n"
        "Yjw7vPe4BnbY6SppSxscW3KoV7GtJLFKIxAXbxsuJMF/rYe3O3w2VKJ1Sug1VDJl\n"
        "/GytwAkSUer84WwP2b07Wn4c5pCnmLslMgXCLkENgi1NnJMhYVOnckxGDZk54hqP\n"
        "9RbLnkkCgYEA/yKuWEvgdzYRYkqpzB0l9ka7Y00CV4Dha9Of6GjQi9i4VCJ/UFVr\n"
        "UlhTo5y0ZzpcDAPcoZf5CFZsD90a/BpQ3YTtdln2MMCL/Kr3QFmetkmDrt+3wYnX\n"
        "sKESfsa2nZdOATRpl1antpwyD4RzsAeOPwBiACj4fkq5iZJBSI0bxrMCgYEA8GFi\n"
        "qAjgKh81/Uai6KWTOW2kX02LEMVRrnZLQ9VPPLGid4KZDDk1/dEfxjjkcyOxX1Ux\n"
        "Klu4W8ZEdZyzPcJrfk7PdopfGOfrhWzkREK9C40H7ou/1jUecq/STPfSOmxh3Y+D\n"
        "ifMNO6z4sQAHx8VaHaxVsJ7SGR/spr0pkZL+NXsCgYEA84rIgBKWB1W+TGRXJzdf\n"
        "yHIGaCjXpm2pQMN3LmP3RrcuZWm0vBt94dHcrR5l+u/zc6iwEDTAjJvqdU4rdyEr\n"
        "tfkwr7v6TNlQB3WvpWanIPyVzfVSNFX/ZWSsAgZvxYjr9ixw6vzWBXOeOb/Gqu7b\n"
        "cvpLkjmJ0wxDhbXtyXKhZA8CgYBZyvcQb+hUs732M4mtQBSD0kohc5TsGdlOQ1AQ\n"
        "McFcmbpnzDghkclyW8jzwdLMk9uxEeDAwuxWE/UEvhlSi6qdzxC+Zifp5NBc0fVe\n"
        "7lMx2mfJGxj5CnSqQLVdHQHB4zSXkAGB6XHbBd0MOUeuvzDPfs2voVQ4IG3FR0oc\n"
        "3/znuwKBgQChZGH3McQcxmLA28aUwOVbWssfXKdDCsiJO+PEXXlL0maO3SbnFn+Q\n"
        "Tyf8oHI5cdP7AbwDSx9bUfRPjg9dKKmATBFr2bn216pjGxK0OjYOCntFTVr0psRB\n"
        "CrKg52Qrq71/2l4V2NLQZU40Dr1bN9V+Ftd9L0pvpCAEAWpIbLXGDw==\n"
        "-----END RSA PRIVATE KEY-----"
    )


@pytest.fixture
def pubkey_data():
    return (
        "-----BEGIN PUBLIC KEY-----\n"
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA75GR6ZTv5JOv90Vq8tKh\n"
        "KC7YQnhDIo2hM0HVziTEk5R4UQBWa0CKytFMbTONY2msEDwX9iA0x7F5Lgj0X8eD\n"
        "4ZMsYqLzqjWMekLC8bjhxc+EuPo9Dygu3mJ2VgRC7XhlFpmdo5NN8J2E7B/CNB3R\n"
        "4hOcMMZNZdi0xLtFoTfwU61UPfFX14mV2laqLbvDEfQLJhUTDeFFV8EN5Z4H1ttL\n"
        "P3sMXJvc3EvM0JiDVj4l1TWFUHHzeFgCA1Im0lv8i7PFrgW7nyMfK9uDSsUmIp7k\n"
        "6ai4tVzwkTmV5PsriP1ju88Lo3MB4/sUmDv/JmlZ9YyzTO3Po8Uz3Aeq9HJWyBWH\n"
        "AQIDAQAB\n"
        "-----END PUBLIC KEY-----"
    )


@pytest.fixture
def hello_world():
    return b"hello, world"


@pytest.fixture
def hello_world_sig():
    return (
        b"\x63\xa0\x70\xd2\xe4\xd4\x6b\x8a\xa2\x59\x27\x5f\x00\x69"
        b"\x1e\x3c\x50\xed\x50\x13\x09\x80\xe3\x47\x4e\x14\xb5\x7c"
        b"\x07\x26\x4e\x20\x74\xea\x0e\xf8\xda\xff\x1e\x57\x8c\x67"
        b"\x76\x73\xaa\xea\x0f\x0a\xe7\xa2\xe3\x88\xfc\x09\x87\x36"
        b"\x01\x3a\xb7\x4c\x40\xe0\xf4\x54\xc5\xf1\xaa\xb2\x1d\x7f"
        b"\xb6\xd3\xa8\xdd\x28\x69\x8b\x88\xe4\x42\x1e\x48\x3e\x1f"
        b"\xe2\x2b\x3c\x7c\x85\x11\xe9\x59\xd7\xf3\xc2\x21\xd3\x55"
        b"\xcb\x9c\x3c\x93\xcc\x20\xdf\x64\x81\xd0\x0d\xbf\x8e\x8d"
        b"\x47\xec\x1d\x9e\x27\xec\x12\xed\x8b\x5f\xd6\x1d\xec\x8d"
        b"\x77\x5a\x58\x8a\x24\xb6\x0f\x12\xb7\x51\xef\x7d\x85\x0f"
        b"\x49\x39\x02\x81\x15\x08\x70\xd6\xe0\x0b\x31\xff\x5f\xf9"
        b"\xd1\x92\x38\x59\x8c\x22\x9c\xbb\xbf\xcf\x85\x34\xe2\x47"
        b"\xf5\xe2\xaa\xb4\x62\x33\x3c\x13\x78\x33\x87\x08\x9e\xb5"
        b"\xbc\x5d\xc1\xbf\x79\x7c\xfa\x5f\x06\x6a\x3b\x17\x40\x09"
        b"\xb9\x09\xbf\x32\xc3\x00\xe2\xbc\x91\x77\x14\xa5\x23\xf5"
        b"\xf5\xf1\x09\x12\x38\xda\x3b\x6a\x82\x81\x7b\x5e\x1c\xcb"
        b"\xaa\x36\x9b\x08\x36\x03\x14\x96\xa3\x31\x39\x59\x16\x75"
        b"\xc9\xb6\x66\x94\x1b\x97\xff\xc8\xa1\xe3\x21\x35\x23\x06"
        b"\x4c\x9b\xf4\xee"
    )


def test_signer(privkey_data, pubkey_data, hello_world, hello_world_sig):
    with pytest.raises(ValueError):
        signer = RSAX931Signer("bogus key data")
    with pytest.raises(ValueError):
        signer = RSAX931Signer(pubkey_data)

    signer = RSAX931Signer(privkey_data)
    with pytest.raises(ValueError):
        signer.sign("x" * 255)  # message too long

    sig = signer.sign(hello_world)
    assert hello_world_sig == sig


def test_verifier(privkey_data, pubkey_data, hello_world, hello_world_sig):
    with pytest.raises(ValueError):
        verifier = RSAX931Verifier("bogus key data")
    with pytest.raises(ValueError):
        verifier = RSAX931Verifier(privkey_data)

    verifier = RSAX931Verifier(pubkey_data)
    with pytest.raises(ValueError):
        verifier.verify("")
    with pytest.raises(ValueError):
        verifier.verify(hello_world_sig + b"junk")

    msg = verifier.verify(hello_world_sig)
    assert hello_world == msg


@pytest.mark.skip_unless_on_windows
def test_find_libcrypto_win32():
    """
    Test _find_libcrypto on Windows hosts.
    """
    lib_path = _find_libcrypto()
    assert "libcrypto" in lib_path


@pytest.mark.skip_unless_on_smartos
def test_find_libcrypto_smartos():
    """
    Test _find_libcrypto on a SmartOS host.
    """
    lib_path = _find_libcrypto()
    assert fnmatch.fnmatch(
        lib_path, os.path.join(os.path.dirname(sys.executable), "libcrypto*")
    )


@pytest.mark.skip_unless_on_sunos
def test_find_libcrypto_sunos():
    """
    Test _find_libcrypto on a Solaris-like host.
    """
    lib_path = _find_libcrypto()
    passed = False
    for i in ("/opt/local/lib/libcrypto.so*", "/opt/tools/lib/libcrypto.so*"):
        if fnmatch.fnmatch(lib_path, i):
            passed = True
            break
    assert passed


@pytest.mark.skip_unless_on_aix
def test_find_libcrypto_aix():
    """
    Test _find_libcrypto on an IBM AIX host.
    """
    lib_path = _find_libcrypto()
    if os.path.isdir("/opt/salt/lib"):
        assert fnmatch.fnmatch(lib_path, "/opt/salt/lib/libcrypto.so*")
    else:
        assert fnmatch.fnmatch(lib_path, "/opt/freeware/lib/libcrypto.so*")


def test_find_libcrypto_with_system_before_catalina():
    """
    Test _find_libcrypto on a pre-Catalina macOS host by simulating not
    finding any other libcryptos and verifying that it defaults to system.
    """
    with patch.object(salt.utils.platform, "is_darwin", lambda: True), patch.object(
        platform, "mac_ver", lambda: ("10.14.2", (), "")
    ), patch.object(glob, "glob", lambda _: []), patch.object(
        sys, "platform", "macosx"
    ):
        lib_path = _find_libcrypto()
        assert lib_path == "/usr/lib/libcrypto.dylib"


def test_find_libcrypto_darwin_catalina():
    """
    Test _find_libcrypto on a macOS Catalina host where there are no custom
    libcryptos and defaulting to the versioned system libraries.
    """
    available = [
        "/usr/lib/libcrypto.0.9.7.dylib",
        "/usr/lib/libcrypto.0.9.8.dylib",
        "/usr/lib/libcrypto.35.dylib",
        "/usr/lib/libcrypto.41.dylib",
        "/usr/lib/libcrypto.42.dylib",
        "/usr/lib/libcrypto.44.dylib",
        "/usr/lib/libcrypto.dylib",
    ]

    def test_glob(pattern):
        return [lib for lib in available if fnmatch.fnmatch(lib, pattern)]

    with patch.object(salt.utils.platform, "is_darwin", lambda: True), patch.object(
        platform, "mac_ver", lambda: ("10.15.2", (), "")
    ), patch.object(sys, "platform", "macosx"), patch.object(glob, "glob", test_glob):
        lib_path = _find_libcrypto()
    assert "/usr/lib/libcrypto.44.dylib" == lib_path


def test_find_libcrypto_darwin_bigsur_packaged():
    """
    Test _find_libcrypto on a Darwin-like macOS host where there isn't a
    lacation returned by ctypes.util.find_library() and the libcrypto
    installation comes from a package manager (ports, brew, salt).
    """
    managed_paths = {
        "salt": "/opt/salt/lib/libcrypto.dylib",
        "brew": "/test/homebrew/prefix/opt/openssl/lib/libcrypto.dylib",
        "port": "/opt/local/lib/libcrypto.dylib",
    }

    saved_getenv = os.getenv

    def mock_getenv(env):
        def test_getenv(var, default=None):
            return env.get(var, saved_getenv(var, default))

        return test_getenv

    def mock_glob(expected_lib):
        def test_glob(pattern):
            if fnmatch.fnmatch(expected_lib, pattern):
                return [expected_lib]
            return []

        return test_glob

    with patch.object(salt.utils.platform, "is_darwin", lambda: True), patch.object(
        platform, "mac_ver", lambda: ("11.2.2", (), "")
    ), patch.object(sys, "platform", "macosx"):
        for package_manager, expected_lib in managed_paths.items():
            if package_manager == "brew":
                env = {"HOMEBREW_PREFIX": "/test/homebrew/prefix"}
            else:
                env = {"HOMEBREW_PREFIX": ""}
            with patch.object(os, "getenv", mock_getenv(env)):
                with patch.object(glob, "glob", mock_glob(expected_lib)):
                    lib_path = _find_libcrypto()

            assert expected_lib == lib_path

        # On Big Sur, there's nothing else to fall back on.
        with patch.object(glob, "glob", lambda _: []):
            with pytest.raises(OSError):
                lib_path = _find_libcrypto()


def test_find_libcrypto_unsupported():
    """
    Ensure that _find_libcrypto works correctly on an unsupported host OS.
    """
    with patch.object(ctypes.util, "find_library", lambda a: None), patch.object(
        glob, "glob", lambda a: []
    ), patch.object(sys, "platform", "unknown"), patch.object(
        salt.utils.platform, "is_darwin", lambda: False
    ), pytest.raises(
        OSError
    ):
        _find_libcrypto()


def test_load_libcrypto():
    """
    Test _load_libcrypto generically.
    """
    lib = _load_libcrypto()
    assert isinstance(lib, ctypes.CDLL)
    # Try to cover both pre and post OpenSSL 1.1.
    assert (
        hasattr(lib, "OpenSSL_version_num")
        or hasattr(lib, "OPENSSL_init_crypto")
        or hasattr(lib, "OPENSSL_no_config")
    )


def test_find_libcrypto_darwin_onedir():
    """
    Test _find_libcrypto on a macOS
    libcryptos and defaulting to the versioned system libraries.
    """
    available = [
        "/usr/lib/libcrypto.0.9.7.dylib",
        "/usr/lib/libcrypto.0.9.8.dylib",
        "/usr/lib/libcrypto.35.dylib",
        "/usr/lib/libcrypto.41.dylib",
        "/usr/lib/libcrypto.42.dylib",
        "/usr/lib/libcrypto.44.dylib",
        "/test/homebrew/prefix/opt/openssl/lib/libcrypto.dylib",
        "/opt/local/lib/libcrypto.dylib",
        "lib/libcrypto.dylib",
    ]

    def test_glob(pattern):
        return [lib for lib in available if fnmatch.fnmatch(lib, pattern)]

    with patch.object(glob, "glob", test_glob), patch.object(
        salt.utils.platform, "is_darwin", lambda: True
    ), patch.object(platform, "mac_ver", lambda: ("10.15.2", (), "")), patch.object(
        sys, "platform", "macosx"
    ):
        lib_path = _find_libcrypto()
    assert "lib/libcrypto.dylib" == lib_path



import pytest

import salt.proxy.panos as panos
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    with patch.dict(panos.DETAILS):
        yield {panos: {"__pillar__": {}}}


@pytest.fixture
def opts():
    return {"proxy": {"proxytype": "panos", "host": "hosturl.com", "apikey": "api_key"}}


def test_init(opts):
    for verify in [True, False, None]:
        opts["proxy"]["verify_ssl"] = verify
        if verify is None:
            opts["proxy"].pop("verify_ssl")
            verify = True
        mock_http = MagicMock(
            return_value={"status": 200, "text": "<data>some_test_data</data>"}
        )
        patch_http = patch.dict(panos.__utils__, {"http.query": mock_http})
        with patch_http:
            panos.init(opts)
        assert mock_http.call_args_list == [
            call(
                "https://hosturl.com/api/",
                data={
                    "type": "op",
                    "cmd": "<show><system><info></info></system></show>",
                    "key": "api_key",
                },
                decode=True,
                decode_type="plain",
                method="POST",
                raise_error=True,
                status=True,
                verify_ssl=verify,
            )
        ]



"""
Splay function calls across targeted minions
"""

import logging
import time

import salt.utils.stringutils

log = logging.getLogger(__name__)

_DEFAULT_SPLAYTIME = 300
_HASH_SIZE = 8192
_HASH_VAL = None


def __init__(opts):
    global _HASH_VAL
    _HASH_VAL = _get_hash()


def _get_hash():
    """
    Jenkins One-At-A-Time Hash Function
    More Info: http://en.wikipedia.org/wiki/Jenkins_hash_function#one-at-a-time
    """
    # Using bitmask to emulate rollover behavior of C unsigned 32 bit int
    bitmask = 0xFFFFFFFF
    h = 0

    for i in bytearray(salt.utils.stringutils.to_bytes(__grains__["id"])):
        h = (h + i) & bitmask
        h = (h + (h << 10)) & bitmask
        h = (h ^ (h >> 6)) & bitmask

    h = (h + (h << 3)) & bitmask
    h = (h ^ (h >> 11)) & bitmask
    h = (h + (h << 15)) & bitmask

    return (h & (_HASH_SIZE - 1)) & bitmask


def _calc_splay(splaytime):
    return int(splaytime * _HASH_VAL / float(_HASH_SIZE))


def execute(opts, data, func, args, kwargs):
    """
    Splay a salt function call execution time across minions over
    a number of seconds (default: 300)

    .. note::
        You *probably* want to use --async here and look up the job results later.
        If you're dead set on getting the output from the CLI command, then make
        sure to set the timeout (with the -t flag) to something greater than the
        splaytime (max splaytime + time to execute job).
        Otherwise, it's very likely that the cli will time out before the job returns.

    CLI Example:

    .. code-block:: bash

        # With default splaytime
        salt --async --module-executors='[splay, direct_call]' '*' pkg.install cowsay version=3.03-8.el6

    .. code-block:: bash

        # With specified splaytime (5 minutes) and timeout with 10 second buffer
        salt -t 310 --module-executors='[splay, direct_call]' --executor-opts='{splaytime: 300}' '*' pkg.version cowsay
    """
    if "executor_opts" in data and "splaytime" in data["executor_opts"]:
        splaytime = data["executor_opts"]["splaytime"]
    else:
        splaytime = opts.get("splaytime", _DEFAULT_SPLAYTIME)
    if splaytime <= 0:
        raise ValueError("splaytime must be a positive integer")
    fun_name = data.get("fun")
    my_delay = _calc_splay(splaytime)
    log.debug("Splay is sleeping %s secs on %s", my_delay, fun_name)

    time.sleep(my_delay)
    return None



"""
Manage Glassfish/Payara server
.. versionadded:: 2016.11.0

Management of glassfish using its RESTful API
You can setup connection parameters like this

.. code-block:: yaml

    - server:
      - ssl: true
      - url: localhost
      - port: 4848
      - user: admin
      - password: changeit
"""

import salt.utils.json
from salt.exceptions import CommandExecutionError

try:
    import requests

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False


def __virtual__():
    """
    Only load if glassfish module is available
    """
    if "glassfish.enum_connector_c_pool" in __salt__ and HAS_LIBS:
        return True
    return (False, "glassfish module could not be loaded")


def _json_to_unicode(data):
    """
    Encode json values in unicode to match that of the API
    """
    ret = {}
    for key, value in data.items():
        if not isinstance(value, str):
            if isinstance(value, dict):
                ret[key] = _json_to_unicode(value)
            else:
                ret[key] = str(value).lower()
        else:
            ret[key] = value
    return ret


def _is_updated(old_conf, new_conf):
    """
    Compare the API results to the current statefile data
    """
    changed = {}

    # Dirty json hacking to get parameters in the same format
    new_conf = _json_to_unicode(
        salt.utils.json.loads(salt.utils.json.dumps(new_conf, ensure_ascii=False))
    )
    old_conf = salt.utils.json.loads(
        salt.utils.json.dumps(old_conf, ensure_ascii=False)
    )

    for key, value in old_conf.items():
        oldval = str(value).lower()
        if key in new_conf:
            newval = str(new_conf[key]).lower()
        if oldval == "null" or oldval == "none":
            oldval = ""
        if key in new_conf and newval != oldval:
            changed[key] = {"old": oldval, "new": newval}
    return changed


def _do_element_present(name, elem_type, data, server=None):
    """
    Generic function to create or update an element
    """
    ret = {"changes": {}, "update": False, "create": False, "error": None}
    try:
        elements = __salt__["glassfish.enum_{}".format(elem_type)]()
    except requests.ConnectionError as error:
        if __opts__["test"]:
            ret["changes"] = {"Name": name, "Params": data}
            ret["create"] = True
            return ret
        else:
            ret["error"] = "Can't connect to the server"
            return ret

    if not elements or name not in elements:
        ret["changes"] = {"Name": name, "Params": data}
        ret["create"] = True
        if not __opts__["test"]:
            try:
                __salt__["glassfish.create_{}".format(elem_type)](
                    name, server=server, **data
                )
            except CommandExecutionError as error:
                ret["error"] = error
                return ret
    elif elements and any(data):
        current_data = __salt__["glassfish.get_{}".format(elem_type)](
            name, server=server
        )
        data_diff = _is_updated(current_data, data)
        if data_diff:
            ret["update"] = True
            ret["changes"] = data_diff
            if not __opts__["test"]:
                try:
                    __salt__["glassfish.update_{}".format(elem_type)](
                        name, server=server, **data
                    )
                except CommandExecutionError as error:
                    ret["error"] = error
    return ret


def _do_element_absent(name, elem_type, data, server=None):
    """
    Generic function to delete an element
    """
    ret = {"delete": False, "error": None}
    try:
        elements = __salt__["glassfish.enum_{}".format(elem_type)]()
    except requests.ConnectionError as error:
        if __opts__["test"]:
            ret["create"] = True
            return ret
        else:
            ret["error"] = "Can't connect to the server"
            return ret

    if elements and name in elements:
        ret["delete"] = True
        if not __opts__["test"]:
            try:
                __salt__["glassfish.delete_{}".format(elem_type)](
                    name, server=server, **data
                )
            except CommandExecutionError as error:
                ret["error"] = error
    return ret


def connection_factory_present(
    name,
    restype="connection_factory",
    description="",
    enabled=True,
    min_size=1,
    max_size=250,
    resize_quantity=2,
    idle_timeout=300,
    wait_timeout=60,
    reconnect_on_failure=False,
    transaction_support="",
    connection_validation=False,
    server=None,
):
    """
    Ensures that the Connection Factory is present

    name
        Name of the connection factory

    restype
        Type of the connection factory, can be either ``connection_factory``,
        ``queue_connection_factory` or ``topic_connection_factory``,
        defaults to ``connection_factory``

    description
        Description of the connection factory

    enabled
        Is the connection factory enabled? defaults to ``true``

    min_size
        Minimum and initial number of connections in the pool, defaults to ``1``

    max_size
        Maximum number of connections that can be created in the pool, defaults to ``250``

    resize_quantity
        Number of connections to be removed when idle_timeout expires, defaults to ``2``

    idle_timeout
        Maximum time a connection can remain idle in the pool, in seconds, defaults to ``300``

    wait_timeout
        Maximum time a caller can wait before timeout, in seconds, defaults to ``60``

    reconnect_on_failure
        Close all connections and reconnect on failure (or reconnect only when used), defaults to ``false``
    transaction_support
        Level of transaction support, can be either ``XATransaction``, ``LocalTransaction`` or ``NoTransaction``

    connection_validation
        Connection validation is required, defaults to ``false``
    """
    ret = {"name": name, "result": None, "comment": None, "changes": {}}

    # Manage parameters
    pool_data = {}
    res_data = {}
    pool_name = "{}-Connection-Pool".format(name)
    if restype == "topic_connection_factory":
        pool_data["connectionDefinitionName"] = "javax.jms.TopicConnectionFactory"
    elif restype == "queue_connection_factory":
        pool_data["connectionDefinitionName"] = "javax.jms.QueueConnectionFactory"
    elif restype == "connection_factory":
        pool_data["connectionDefinitionName"] = "javax.jms.ConnectionFactory"
    else:
        ret["result"] = False
        ret["comment"] = "Invalid restype"
        return ret
    pool_data["description"] = description
    res_data["description"] = description
    res_data["enabled"] = enabled
    res_data["poolName"] = pool_name
    pool_data["steadyPoolSize"] = min_size
    pool_data["maxPoolSize"] = max_size
    pool_data["poolResizeQuantity"] = resize_quantity
    pool_data["idleTimeoutInSeconds"] = idle_timeout
    pool_data["maxWaitTimeInMillis"] = wait_timeout * 1000
    pool_data["failAllConnections"] = reconnect_on_failure
    if transaction_support:
        if transaction_support == "xa_transaction":
            pool_data["transactionSupport"] = "XATransaction"
        elif transaction_support == "local_transaction":
            pool_data["transactionSupport"] = "LocalTransaction"
        elif transaction_support == "no_transaction":
            pool_data["transactionSupport"] = "NoTransaction"
        else:
            ret["result"] = False
            ret["comment"] = "Invalid transaction_support"
            return ret
    pool_data["isConnectionValidationRequired"] = connection_validation

    pool_ret = _do_element_present(pool_name, "connector_c_pool", pool_data, server)
    res_ret = _do_element_present(name, "connector_resource", res_data, server)

    if not pool_ret["error"] and not res_ret["error"]:
        if not __opts__["test"]:
            ret["result"] = True

        if pool_ret["create"] or res_ret["create"]:
            ret["changes"]["pool"] = pool_ret["changes"]
            ret["changes"]["resource"] = res_ret["changes"]
            if __opts__["test"]:
                ret["comment"] = "Connection factory set to be created"
            else:
                ret["comment"] = "Connection factory created"
        elif pool_ret["update"] or res_ret["update"]:
            ret["changes"]["pool"] = pool_ret["changes"]
            ret["changes"]["resource"] = res_ret["changes"]
            if __opts__["test"]:
                ret["comment"] = "Connection factory set to be updated"
            else:
                ret["comment"] = "Connection factory updated"
        else:
            ret["result"] = True
            ret["changes"] = {}
            ret["comment"] = "Connection factory is already up-to-date"
    else:
        ret["result"] = False
        ret["comment"] = "ERROR: {} // {}".format(pool_ret["error"], res_ret["error"])

    return ret


def connection_factory_absent(name, both=True, server=None):
    """
    Ensures the transaction factory is absent.

    name
        Name of the connection factory

    both
        Delete both the pool and the resource, defaults to ``true``
    """
    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    pool_name = "{}-Connection-Pool".format(name)
    pool_ret = _do_element_absent(
        pool_name, "connector_c_pool", {"cascade": both}, server
    )

    if not pool_ret["error"]:
        if __opts__["test"] and pool_ret["delete"]:
            ret["comment"] = "Connection Factory set to be deleted"
        elif pool_ret["delete"]:
            ret["result"] = True
            ret["comment"] = "Connection Factory deleted"
        else:
            ret["result"] = True
            ret["comment"] = "Connection Factory doesn't exist"
    else:
        ret["result"] = False
        ret["comment"] = "Error: {}".format(pool_ret["error"])
    return ret


def destination_present(
    name, physical, restype="queue", description="", enabled=True, server=None
):
    """
    Ensures that the JMS Destination Resource (queue or topic) is present

    name
        The JMS Queue/Topic name

    physical
        The Physical destination name

    restype
        The JMS Destination resource type, either ``queue`` or ``topic``, defaults is ``queue``

    description
        A description of the resource

    enabled
        Defaults to ``True``
    """
    ret = {"name": name, "result": None, "comment": None, "changes": {}}

    params = {}
    # Set parameters dict
    if restype == "queue":
        params["resType"] = "javax.jms.Queue"
        params["className"] = "com.sun.messaging.Queue"
    elif restype == "topic":
        params["resType"] = "javax.jms.Topic"
        params["className"] = "com.sun.messaging.Topic"
    else:
        ret["result"] = False
        ret["comment"] = "Invalid restype"
        return ret
    params["properties"] = {"Name": physical}
    params["description"] = description
    params["enabled"] = enabled

    jms_ret = _do_element_present(name, "admin_object_resource", params, server)
    if not jms_ret["error"]:
        if not __opts__["test"]:
            ret["result"] = True
        if jms_ret["create"] and __opts__["test"]:
            ret["comment"] = "JMS Queue set to be created"
        elif jms_ret["create"]:
            ret["changes"] = jms_ret["changes"]
            ret["comment"] = "JMS queue created"
        elif jms_ret["update"] and __opts__["test"]:
            ret["comment"] = "JMS Queue set to be updated"
        elif jms_ret["update"]:
            ret["changes"] = jms_ret["changes"]
            ret["comment"] = "JMS Queue updated"
        else:
            ret["result"] = True
            ret["comment"] = "JMS Queue already up-to-date"
    else:
        ret["result"] = False
        ret["comment"] = "Error from API: {}".format(jms_ret["error"])
    return ret


def destination_absent(name, server=None):
    """
    Ensures that the JMS Destination doesn't exists

    name
        Name of the JMS Destination
    """
    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    jms_ret = _do_element_absent(name, "admin_object_resource", {}, server)
    if not jms_ret["error"]:
        if __opts__["test"] and jms_ret["delete"]:
            ret["comment"] = "JMS Queue set to be deleted"
        elif jms_ret["delete"]:
            ret["result"] = True
            ret["comment"] = "JMS Queue deleted"
        else:
            ret["result"] = True
            ret["comment"] = "JMS Queue doesn't exist"
    else:
        ret["result"] = False
        ret["comment"] = "Error: {}".format(jms_ret["error"])
    return ret


def jdbc_datasource_present(
    name,
    description="",
    enabled=True,
    restype="datasource",
    vendor="mysql",
    sql_url="",
    sql_user="",
    sql_password="",
    min_size=8,
    max_size=32,
    resize_quantity=2,
    idle_timeout=300,
    wait_timeout=60,
    non_transactional=False,
    transaction_isolation="",
    isolation_guaranteed=True,
    server=None,
):
    """
    Ensures that the JDBC Datasource exists

    name
        Name of the datasource

    description
        Description of the datasource

    enabled
        Is the datasource enabled? defaults to ``true``

    restype
        Resource type, can be ``datasource``, ``xa_datasource``,
        ``connection_pool_datasource`` or ``driver``, defaults to ``datasource``

    vendor
        SQL Server type, currently supports ``mysql``,
        ``postgresql`` and ``mssql``, defaults to ``mysql``

    sql_url
        URL of the server in jdbc form

    sql_user
        Username for the server

    sql_password
        Password for that username

    min_size
        Minimum and initial number of connections in the pool, defaults to ``8``

    max_size
        Maximum number of connections that can be created in the pool, defaults to ``32``

    resize_quantity
        Number of connections to be removed when idle_timeout expires, defaults to ``2``

    idle_timeout
        Maximum time a connection can remain idle in the pool, in seconds, defaults to ``300``

    wait_timeout
        Maximum time a caller can wait before timeout, in seconds, defaults to ``60``

    non_transactional
        Return non-transactional connections

    transaction_isolation
        Defaults to the JDBC driver default

    isolation_guaranteed
        All connections use the same isolation level
    """
    ret = {"name": name, "result": None, "comment": None, "changes": {}}

    # Manage parameters
    res_name = "jdbc/{}".format(name)
    pool_data = {}
    pool_data_properties = {}
    res_data = {}
    if restype == "datasource":
        pool_data["resType"] = "javax.sql.DataSource"
    elif restype == "xa_datasource":
        pool_data["resType"] = "javax.sql.XADataSource"
    elif restype == "connection_pool_datasource":
        pool_data["resType"] = "javax.sql.ConnectionPoolDataSource"
    elif restype == "driver":
        pool_data["resType"] = "javax.sql.Driver"

    datasources = {}
    datasources["mysql"] = {
        "driver": "com.mysql.jdbc.Driver",
        "datasource": "com.mysql.jdbc.jdbc2.optional.MysqlDataSource",
        "xa_datasource": "com.mysql.jdbc.jdbc2.optional.MysqlXADataSource",
        "connection_pool_datasource": (
            "com.mysql.jdbc.jdbc2.optional.MysqlConnectionPoolDataSource"
        ),
    }
    datasources["postgresql"] = {
        "driver": "org.postgresql.Driver",
        "datasource": "org.postgresql.ds.PGSimpleDataSource",
        "xa_datasource": "org.postgresql.xa.PGXADataSource",
        "connection_pool_datasource": "org.postgresql.ds.PGConnectionPoolDataSource",
    }
    datasources["mssql"] = {
        "driver": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
        "datasource": "com.microsoft.sqlserver.jdbc.SQLServerDataSource",
        "xa_datasource": "com.microsoft.sqlserver.jdbc.SQLServerXADataSource",
        "connection_pool_datasource": (
            "com.microsoft.sqlserver.jdbc.SQLServerConnectionPoolDataSource"
        ),
    }

    if restype == "driver":
        pool_data["driverClassname"] = datasources[vendor]["driver"]
    else:
        pool_data["datasourceClassname"] = datasources[vendor][restype]

    pool_data_properties["url"] = sql_url
    pool_data_properties["user"] = sql_user
    pool_data_properties["password"] = sql_password
    pool_data["properties"] = pool_data_properties
    pool_data["description"] = description
    res_data["description"] = description
    res_data["poolName"] = name
    res_data["enabled"] = enabled
    pool_data["steadyPoolSize"] = min_size
    pool_data["maxPoolSize"] = max_size
    pool_data["poolResizeQuantity"] = resize_quantity
    pool_data["idleTimeoutInSeconds"] = idle_timeout
    pool_data["maxWaitTimeInMillis"] = wait_timeout * 1000
    pool_data["nonTransactionalConnections"] = non_transactional
    pool_data["transactionIsolationLevel"] = transaction_isolation
    pool_data["isIsolationLevelGuaranteed"] = isolation_guaranteed

    pool_ret = _do_element_present(name, "jdbc_connection_pool", pool_data, server)
    res_ret = _do_element_present(res_name, "jdbc_resource", res_data, server)

    if not pool_ret["error"] and not res_ret["error"]:
        if not __opts__["test"]:
            ret["result"] = True

        if pool_ret["create"] or res_ret["create"]:
            ret["changes"]["pool"] = pool_ret["changes"]
            ret["changes"]["resource"] = res_ret["changes"]
            if __opts__["test"]:
                ret["comment"] = "JDBC Datasource set to be created"
            else:
                ret["comment"] = "JDBC Datasource created"
        elif pool_ret["update"] or res_ret["update"]:
            ret["changes"]["pool"] = pool_ret["changes"]
            ret["changes"]["resource"] = res_ret["changes"]
            if __opts__["test"]:
                ret["comment"] = "JDBC Datasource set to be updated"
            else:
                ret["comment"] = "JDBC Datasource updated"
        else:
            ret["result"] = True
            ret["changes"] = {}
            ret["comment"] = "JDBC Datasource is already up-to-date"
    else:
        ret["result"] = False
        ret["comment"] = "ERROR: {} // {}".format(pool_ret["error"], res_ret["error"])

    return ret


def jdbc_datasource_absent(name, both=True, server=None):
    """
    Ensures the JDBC Datasource doesn't exists

    name
        Name of the datasource
    both
        Delete both the pool and the resource, defaults to ``true``
    """
    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    pool_ret = _do_element_absent(
        name, "jdbc_connection_pool", {"cascade": both}, server
    )

    if not pool_ret["error"]:
        if __opts__["test"] and pool_ret["delete"]:
            ret["comment"] = "JDBC Datasource set to be deleted"
        elif pool_ret["delete"]:
            ret["result"] = True
            ret["comment"] = "JDBC Datasource deleted"
        else:
            ret["result"] = True
            ret["comment"] = "JDBC Datasource doesn't exist"
    else:
        ret["result"] = False
        ret["comment"] = "Error: {}".format(pool_ret["error"])
    return ret


def system_properties_present(server=None, **kwargs):
    """
    Ensures that the system properties are present

    properties
        The system properties
    """
    ret = {"name": "", "result": None, "comment": None, "changes": {}}

    del kwargs["name"]
    try:
        data = __salt__["glassfish.get_system_properties"](server=server)
    except requests.ConnectionError as error:
        if __opts__["test"]:
            ret["changes"] = kwargs
            ret["result"] = None
            return ret
        else:
            ret["error"] = "Can't connect to the server"
            return ret

    ret["changes"] = {"data": data, "kwargs": kwargs}
    if not data == kwargs:
        data.update(kwargs)
        if not __opts__["test"]:
            try:
                __salt__["glassfish.update_system_properties"](data, server=server)
                ret["changes"] = kwargs
                ret["result"] = True
                ret["comment"] = "System properties updated"
            except CommandExecutionError as error:
                ret["comment"] = error
                ret["result"] = False
        else:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["coment"] = "System properties would have been updated"
    else:
        ret["changes"] = {}
        ret["result"] = True
        ret["comment"] = "System properties are already up-to-date"
    return ret


def system_properties_absent(name, server=None):
    """
    Ensures that the system property doesn't exists

    name
        Name of the system property
    """
    ret = {"name": "", "result": None, "comment": None, "changes": {}}

    try:
        data = __salt__["glassfish.get_system_properties"](server=server)
    except requests.ConnectionError as error:
        if __opts__["test"]:
            ret["changes"] = {"Name": name}
            ret["result"] = None
            return ret
        else:
            ret["error"] = "Can't connect to the server"
            return ret

    if name in data:
        if not __opts__["test"]:
            try:
                __salt__["glassfish.delete_system_properties"](name, server=server)
                ret["result"] = True
                ret["comment"] = "System properties deleted"
            except CommandExecutionError as error:
                ret["comment"] = error
                ret["result"] = False
        else:
            ret["result"] = None
            ret["comment"] = "System properties would have been deleted"
        ret["changes"] = {"Name": name}
    else:
        ret["result"] = True
        ret["comment"] = "System properties are already absent"
    return ret



import os

import salt.modules.file as file_
import salt.modules.heat
import salt.modules.win_file as win_file
import salt.states.heat as heat
import salt.utils.platform
import salt.utils.win_dacl as dacl
import tests.unit.modules.test_heat
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class HeatTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.heat
    """

    def setup_loader_modules(self):
        return {
            heat: {
                "_auth": tests.unit.modules.test_heat.MockClient,
                "__opts__": {"test": False},
            },
            salt.modules.heat: {"_auth": tests.unit.modules.test_heat.MockClient},
            file_: {
                "__opts__": {
                    "hash_type": "sha256",
                    "cachedir": os.path.join(RUNTIME_VARS.TMP, "rootdir", "cache"),
                    "test": False,
                },
                "__salt__": {
                    "config.option": MagicMock(
                        return_value={"obfuscate_templates": False}
                    ),
                    "config.backup_mode": MagicMock(return_value=False),
                },
            },
            win_file: {
                "__utils__": {"dacl.check_perms": salt.utils.win_dacl.check_perms}
            },
            dacl: {"__opts__": {"test": False}},
        }

    def setUp(self):
        self.patch_check = patch("salt.modules.file.check_perms", file_.check_perms)
        if salt.utils.platform.is_windows():
            self.patch_check = patch(
                "salt.modules.file.check_perms", win_file.check_perms
            )

    def test_heat_deployed(self):
        """
        Test salt.states.heat.deployed method
        """
        exp_ret = {
            "name": ("mystack",),
            "comment": "Created stack 'mystack'.",
            "changes": {"stack_name": "mystack", "comment": "Create stack"},
            "result": True,
        }

        patch_heat = patch.dict(
            heat.__salt__,
            {
                "heat.show_stack": MagicMock(return_value={"result": False}),
                "heat.create_stack": salt.modules.heat.create_stack,
            },
        )

        patch_file = patch.dict(
            "salt.modules.heat.__salt__",
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": file_.manage_file,
            },
        )

        patch_create = patch(
            "salt.modules.heat.create_stack",
            MagicMock(
                return_value={"result": True, "comment": "Created stack 'mystack'."}
            ),
        )

        with patch_heat, patch_file, patch_create, self.patch_check:
            ret = heat.deployed(
                name="mystack",
                profile="openstack1",
                template=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
                poll=0,
            )
        assert ret == exp_ret

    def test_heat_deployed_environment(self):
        """
        Test salt.states.heat.deployed method
        with environment set
        """
        exp_ret = {
            "name": ("mystack",),
            "comment": "Created stack 'mystack'.",
            "changes": {"stack_name": "mystack", "comment": "Create stack"},
            "result": True,
        }

        patch_heat = patch.dict(
            heat.__salt__,
            {
                "heat.show_stack": MagicMock(return_value={"result": False}),
                "heat.create_stack": salt.modules.heat.create_stack,
            },
        )

        patch_file = patch.dict(
            "salt.modules.heat.__salt__",
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": file_.manage_file,
            },
        )

        patch_create = patch(
            "salt.modules.heat.create_stack",
            MagicMock(
                return_value={"result": True, "comment": "Created stack 'mystack'."}
            ),
        )

        with patch_heat, patch_file, patch_create, self.patch_check:
            ret = heat.deployed(
                name="mystack",
                profile="openstack1",
                template=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
                poll=0,
                environment=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-env.yml"
                ),
            )
        assert ret == exp_ret

    def test_heat_deployed_environment_error(self):
        """
        Test salt.states.heat.deployed method
        with environment set and there is an error
        reading the environment file.
        """
        exp_ret = {
            "name": ("mystack",),
            "comment": "Error parsing template Template format version not found.",
            "changes": {"stack_name": "mystack", "comment": "Create stack"},
            "result": False,
        }

        patch_heat = patch.dict(
            heat.__salt__,
            {
                "heat.show_stack": MagicMock(return_value={"result": False}),
                "heat.create_stack": salt.modules.heat.create_stack,
            },
        )

        patch_file = patch.dict(
            "salt.modules.heat.__salt__",
            {
                "file.get_managed": file_.get_managed,
                "file.manage_file": MagicMock(
                    side_effect=[{"result": True}, {"result": False}]
                ),
            },
        )

        patch_create = patch(
            "salt.modules.heat.create_stack",
            MagicMock(
                return_value={"result": True, "comment": "Created stack 'mystack'."}
            ),
        )

        with patch_heat, patch_file, patch_create, self.patch_check:
            ret = heat.deployed(
                name="mystack",
                profile="openstack1",
                template=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-template.yml"
                ),
                poll=0,
                environment=os.path.join(
                    RUNTIME_VARS.BASE_FILES, "templates", "heat-env.yml"
                ),
            )
        assert ret == exp_ret



"""
Tests for _clean_dir function
"""
import pytest

import salt.states.file as file

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_normal():
    expected = []
    result = file._clean_dir(
        root=r"/tmp/parent",
        keep=[r"/tmp/parent/meh-1.txt", r"/tmp/parent/meh-2.txt"],
        exclude_pat=None,
    )
    assert result == expected


def test_win_forward_slash():
    expected = []
    result = file._clean_dir(
        root=r"C:/test/parent",
        keep=[r"C:/test/parent/meh-1.txt", r"C:/test/parent/meh-2.txt"],
        exclude_pat=None,
    )
    assert result == expected


def test_win_forward_slash_mixed_case():
    expected = []
    result = file._clean_dir(
        root=r"C:/test/parent",
        keep=[r"C:/test/parent/meh-1.txt", r"C:/test/Parent/Meh-2.txt"],
        exclude_pat=None,
    )
    assert result == expected


def test_win_back_slash():
    expected = []
    result = file._clean_dir(
        root=r"C:\test\parent",
        keep=[r"C:\test\parent\meh-1.txt", r"C:\test\parent\meh-2.txt"],
        exclude_pat=None,
    )
    assert result == expected


def test_win_back_slash_mixed_cased():
    expected = []
    result = file._clean_dir(
        root=r"C:\test\parent",
        keep=[r"C:\test\parent\meh-1.txt", r"C:\test\Parent\Meh-2.txt"],
        exclude_pat=None,
    )
    assert result == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.win_ntp as win_ntp
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_ntp: {}}


def test_set_servers():
    """
    Test if it set Windows to use a list of NTP servers
    """
    # Windows Time (W32Time) service is not started
    # Windows Time (W32Time) service fails to start
    mock_service = MagicMock(return_value=False)
    with patch.dict(
        win_ntp.__salt__,
        {"service.status": mock_service, "service.start": mock_service},
    ):
        assert not win_ntp.set_servers("pool.ntp.org")

    # Windows Time service is running
    # Fail to set NTP servers
    mock_service = MagicMock(return_value=True)
    mock_cmd = MagicMock(
        side_effect=[
            "Failure",
            "Failure",
            "Failure",
            "NtpServer: time.windows.com,0x8",
        ]
    )
    with patch.dict(
        win_ntp.__salt__, {"service.status": mock_service, "cmd.run": mock_cmd}
    ):
        assert not win_ntp.set_servers("pool.ntp.org")

    # Windows Time service is running
    # Successfully set NTP servers
    mock_cmd = MagicMock(
        side_effect=["Success", "Success", "Success", "NtpServer: pool.ntp.org"]
    )
    with patch.dict(
        win_ntp.__salt__,
        {
            "service.status": mock_service,
            "service.restart": mock_service,
            "cmd.run": mock_cmd,
        },
    ):
        assert win_ntp.set_servers("pool.ntp.org")


def test_get_servers():
    """
    Test if it get list of configured NTP servers
    """
    mock_cmd = MagicMock(side_effect=["", "NtpServer: SALT", "NtpServer"])
    with patch.dict(win_ntp.__salt__, {"cmd.run": mock_cmd}):
        assert not win_ntp.get_servers()

        assert win_ntp.get_servers() == ["SALT"]

        assert not win_ntp.get_servers()



#
# Copyright 2016 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import fnmatch
import os.path
import re

import salt.utils.stringutils
from salt.exceptions import CommandExecutionError


class InputSanitizer:
    @staticmethod
    def trim(value):
        """
        Raise an exception if value is empty. Otherwise strip it down.
        :param value:
        :return:
        """
        value = (value or "").strip()
        if not value:
            raise CommandExecutionError("Empty value during sanitation")

        return str(value)

    @staticmethod
    def filename(value):
        """
        Remove everything that would affect paths in the filename

        :param value:
        :return:
        """
        return re.sub(
            "[^a-zA-Z0-9.-_ ]", "", os.path.basename(InputSanitizer.trim(value))
        )

    @staticmethod
    def hostname(value):
        """
        Clean value for RFC1123.

        :param value:
        :return:
        """
        return re.sub(r"[^a-zA-Z0-9.-]", "", InputSanitizer.trim(value)).strip(".")

    id = hostname


clean = InputSanitizer()


def mask_args_value(data, mask):
    """
    Mask a line in the data, which matches "mask".

    This can be used for cases where values in your roster file may contain
    sensitive data such as IP addresses, passwords, user names, etc.

    Note that this works only when ``data`` is a single string (i.e. when the
    data in the roster is formatted as ``key: value`` pairs in YAML syntax).

    :param data: String data, already rendered.
    :param mask: Mask that matches a single line

    :return:
    """
    if not mask:
        return data

    out = []
    for line in data.split(os.linesep):
        if fnmatch.fnmatch(line.strip(), mask) and ":" in line:
            key, value = line.split(":", 1)
            out.append(
                "{}: {}".format(
                    salt.utils.stringutils.to_unicode(key.strip()), "** hidden **"
                )
            )
        else:
            out.append(line)

    return "\n".join(out)



"""
    :codeauthor: Alexander Pyatkin <asp@thexyz.net>
"""


import pytest

import salt.states.webutil as htpasswd
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {htpasswd: {"__opts__": {"test": False}}}


def test_user_exists_already():
    """
    Test if it returns True when user already exists in htpasswd file
    """

    mock = MagicMock(return_value={"retcode": 0})
    mock_file_exists = MagicMock(return_value={"retcode": 1})

    with patch.dict(
        htpasswd.__salt__, {"file.grep": mock, "file.file_exists": mock_file_exists}
    ):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": True,
            "comment": "User already known",
            "changes": {},
        }
        assert ret == expected


def test_new_user_success():
    """
    Test if it returns True when new user is added to htpasswd file
    """

    mock_file_exists = MagicMock(return_value={"retcode": 1})
    mock_grep = MagicMock(return_value={"retcode": 1})
    mock_useradd = MagicMock(return_value={"retcode": 0, "stderr": "Success"})

    with patch.dict(
        htpasswd.__salt__,
        {
            "file.grep": mock_grep,
            "file.file_exists": mock_file_exists,
            "webutil.useradd": mock_useradd,
        },
    ):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": True,
            "comment": "Success",
            "changes": {"larry": True},
        }
        assert ret == expected


def test_new_user_error():
    """
    Test if it returns False when adding user to htpasswd failed
    """

    mock_file_exists = MagicMock(return_value={"retcode": 1})
    mock_grep = MagicMock(return_value={"retcode": 1})
    mock_useradd = MagicMock(return_value={"retcode": 1, "stderr": "Error"})

    with patch.dict(
        htpasswd.__salt__,
        {
            "file.grep": mock_grep,
            "file.file_exists": mock_file_exists,
            "webutil.useradd": mock_useradd,
        },
    ):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": False,
            "comment": "Error",
            "changes": {},
        }
        assert ret == expected


def test_no_htpasswd_file_and_user_doesnt_exist(tmp_path):
    """
    Test if .htpasswd does not exist, it returns that the user does not
    exist as opposed to erroring out
    """
    htpasswd_file = tmp_path / ".htpasswd"
    assert not htpasswd_file.is_file()

    mock_file_exists = MagicMock(return_value={"retcode": 1})
    mock_grep = MagicMock(return_value={"retcode": 1})
    mock_useradd = MagicMock(return_value={"retcode": 0, "stderr": "Success"})

    with patch.dict(
        htpasswd.__salt__,
        {
            "file.grep": mock_grep,
            "file.file_exists": mock_file_exists,
            "webutil.useradd": mock_useradd,
        },
    ):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": True,
            "comment": "Success",
            "changes": {"larry": True},
        }
        assert ret == expected



"""
Provide authentication using YubiKey.

.. versionadded:: 2015.5.0

:depends: yubico-client Python module

To get your YubiKey API key you will need to visit the website below.

https://upgrade.yubico.com/getapikey/

The resulting page will show the generated Client ID (aka AuthID or API ID)
and the generated API key (Secret Key). Make a note of both and use these
two values in your /etc/salt/master configuration.

  /etc/salt/master

  .. code-block:: yaml

    yubico_users:
      damian:
        id: 12345
        key: ABCDEFGHIJKLMNOPQRSTUVWXYZ


  .. code-block:: yaml

    external_auth:
      yubico:
        damian:
          - test.*


Please wait five to ten minutes after generating the key before testing so that
the API key will be updated on all the YubiCloud servers.

"""


import logging

log = logging.getLogger(__name__)

try:
    from yubico_client import Yubico, yubico_exceptions

    HAS_YUBICO = True
except ImportError:
    HAS_YUBICO = False


def __get_yubico_users(username):
    """
    Grab the YubiKey Client ID & Secret Key
    """
    user = {}

    try:
        if __opts__["yubico_users"].get(username, None):
            (user["id"], user["key"]) = list(
                __opts__["yubico_users"][username].values()
            )
        else:
            return None
    except KeyError:
        return None

    return user


def auth(username, password):
    """
    Authenticate against yubico server
    """
    _cred = __get_yubico_users(username)

    client = Yubico(_cred["id"], _cred["key"])

    try:
        return client.verify(password)
    except yubico_exceptions.StatusCodeError as e:
        log.info("Unable to verify YubiKey `%s`", e)
        return False


def groups(username, *args, **kwargs):
    return False


if __name__ == "__main__":
    __opts__ = {"yubico_users": {"damian": {"id": "12345", "key": "ABC123"}}}

    if auth("damian", "OPT"):
        print("Authenticated")
    else:
        print("Failed to authenticate")
